Require Import stalmarck.
Require Import LetP.
Require Import Option.
Section algo.
Variable Mem:Set.
Variable memF:Mem ->State.
Variable add:rZ -> rZ -> Mem ->(Mem * bool) * (list rZ).
Hypothesis appSym:(L1, L2:(list rZ))(app L1 L2)=(app L2 L1).

Local getMemBool := (fst Mem * bool (list rZ)).

Local getMem := [x:(Mem * bool) * (list rZ)]<Mem, bool> Fst ((getMemBool x)).

Local getBool := [x:(Mem * bool) * (list rZ)]<Mem, bool> Snd ((getMemBool x)).

Local getList := (snd Mem * bool (list rZ)).
Variable eval:Mem -> rZ ->rZ.

Local eqMem := [M1, M2:Mem] (c:rZ)(eval M1 c)=(eval M2 c).
Require Import Relation_Definitions.

Theorem eqMemSym: (symmetric ? eqMem).
Red; Unfold eqMem; Auto.
Qed.
Hints Immediate eqMemSym.

Theorem eqMemRef: (M:Mem)(eqMem M M).
Red; Auto.
Qed.
Hints Resolve eqMemRef.

Theorem eqMemTrans: (transitive Mem eqMem).
Red; Unfold eqMem; Auto.
Intros x y z H' H'0 c; Rewrite H'; Auto.
Qed.

Local eqExceptOn :=
   [M1, M2:Mem] [L:(list rZ)] (c:rZ) ~ (In c L) -> ~ (In (rZComp c) L) ->
   (eval M1 c)=(eval M2 c).

Theorem eqExceptOnRef:
  (M1, M2:Mem) (L:(list rZ)) (eqMem M1 M2) ->(eqExceptOn M1 M2 L).
Unfold eqMem; Intros M1 M2 L0 H'; Red; Auto.
Qed.
Hints Resolve eqExceptOnRef.

Theorem eqExceptOnrTrans:
  (M1, M2, M3:Mem) (L:(list rZ)) (eqExceptOn M1 M2 L) -> (eqExceptOn M2 M3 L) ->
  (eqExceptOn M1 M3 L).
Unfold eqExceptOn.
Intros M1 M2 M3 L0 H' H'0 c H'1 H'2; Rewrite H'; Auto.
Qed.

Theorem eqExceptOnIncl:
  (M1, M2:Mem) (L1, L2:(list rZ)) (eqExceptOn M1 M2 L1) -> (incl L1 L2) ->
  (eqExceptOn M1 M2 L2).
Unfold eqExceptOn.
Intros M1 M2 L1 L2 H' H'0 c H'1 H'2; Rewrite H'; Auto.
Qed.

Theorem eqExceptOnSym:
  (M1, M2:Mem) (L:(list rZ)) (eqExceptOn M1 M2 L) ->(eqExceptOn M2 M1 L).
Unfold eqExceptOn; Intros M1 M2 L0 H'; Auto.
Intros c H'0 H'1.
Rewrite <- H'; Auto.
Qed.
Hints Immediate eqExceptOnSym.
Hypothesis evalrZTrue:(M:Mem)(eval M rZTrue)=rZTrue.
Hypothesis evalrZComp:(M:Mem) (a:rZ)(eval M (rZComp a))=(rZComp (eval M a)).
Hypothesis evalCorrect:
           (a, b:rZ) (M:Mem) (eqStateRz (memF M) a b) ->(eval M a)=(eval M b).
Hypothesis evalEq:(a:rZ) (M:Mem)(eqStateRz (memF M) a (eval M a)).
Hypothesis addCorrectNotContrEq:
           (M:Mem) (a, b:rZ) (getBool (add a b M))=false ->
           (eqState (addEq (a, b) (memF M)) (memF (getMem (add a b M)))).
Hypothesis addCorrectContrEval:
           (M:Mem) (a, b:rZ) (getBool (add a b M))=true ->
           (eval M a)=(rZComp (eval M b)).
Hypothesis addCorrectNotContrNotEval:
           (M:Mem) (a, b:rZ) (getBool (add a b M))=false ->
           ~ (eval M a)=(rZComp (eval M b)).
Hypothesis addCorrectEval:
           (M:Mem) (a, b:rZ)
           (eqExceptOn (getMem (add a b M)) M (getList (add a b M))).
Variable L:(list triplet).
Variable varList:rZ ->(list triplet).
Hypothesis varListCorrect:
           (a:rZ) (t:triplet) (inTriplet a t) ->(In t (varList a)).
Hypothesis varListCorrectIn:(a:rZ) (t:triplet) (In t (varList a)) ->(In t L).
Variable ltMem:Mem -> Mem ->Prop.
Hypothesis ltMemEqComp:
           (M1, M2, M3, M4:Mem)
           (eqMem M1 M2) -> (eqMem M3 M4) -> (ltMem M1 M3) ->(ltMem M2 M4).
Hypothesis ltWf:(well_founded Mem ltMem).
Require Import Relation_Definitions.
Hypothesis lttrans:(transitive Mem ltMem).

Local ltWithList :=
   [M1, M2:Mem] [L:(list rZ)]
      Cases L of
          nil => (eqMem M1 M2)
         | _ => (ltMem M1 M2)
      end.
Hypothesis ltWithEq:
           (M1, M2, M3, M4:Mem)
           (L:(list rZ))
           (eqMem M1 M2) -> (eqMem M3 M4) -> (ltWithList M1 M3 L) ->
           (ltWithList M2 M4 L).
Hypothesis addCorrectlt:
           (M:Mem) (a, b:rZ)
           (ltWithList (getMem (add a b M)) M (getList (add a b M))).
Variable reset:Mem -> Mem -> (list rZ) ->Mem.
Hypothesis resetClearIn:
           (M1, M2:Mem) (L:(list rZ)) (c:rZ) (In c L) \/ (In (rZComp c) L) ->
           (eval (reset M1 M2 L) c)=(eval M2 c).
Hypothesis resetClearNIn:
           (M1, M2:Mem)
           (L:(list rZ)) (c:rZ) ~ (In c L) -> ~ (In (rZComp c) L) ->
           (eval (reset M1 M2 L) c)=(eval M1 c).
Variable inter:Mem -> Mem -> Mem -> (list rZ) -> (list rZ) ->Mem * (list rZ).
Hypothesis interlt:
           (M1, M2, M3:Mem) (L2, L3:(list rZ))
           (ltWithList
              (Fst (inter M1 M2 M3 L2 L3)) M1 (Snd (inter M1 M2 M3 L2 L3))).
Hypothesis interCorrect:
           (M1, M2, M3:Mem)
           (L2, L3:(list rZ)) (eqExceptOn M1 M2 L2) -> (eqExceptOn M1 M3 L3) ->
           (eqState
              (memF (Fst (inter M1 M2 M3 L2 L3)))
              (interState (memF M2) (memF M3))).
Hypothesis interCorrectIncL:
           (M1, M2, M3:Mem)
           (L2, L3:(list rZ)) (eqExceptOn M1 M2 L2) -> (eqExceptOn M1 M3 L3) ->
           (incl (Snd (inter M1 M2 M3 L2 L3)) L2).
Hypothesis interCorrectIncR:
           (M1, M2, M3:Mem)
           (L2, L3:(list rZ)) (eqExceptOn M1 M2 L2) -> (eqExceptOn M1 M3 L3) ->
           (incl (Snd (inter M1 M2 M3 L2 L3)) L3).
Hypothesis interCorrectEq:
           (M1, M2, M3:Mem)
           (L2, L3:(list rZ)) (eqExceptOn M1 M2 L2) -> (eqExceptOn M1 M3 L3) ->
           (eqExceptOn
              M1 (Fst (inter M1 M2 M3 L2 L3)) (Snd (inter M1 M2 M3 L2 L3))).

Definition memFx: (Mem * bool) * (list rZ) ->State :=
   [L:?]
      Cases L of
          (pair (pair M b) L) => Cases b of
                                     true =>
                                       (cons
                                          ((rZPlus zero), (rZMinus zero))
                                          (nil ?))
                                    | false => (memF M)
                                 end
      end.

Theorem evalCorrectComp:
  (a, b:rZ) (M:Mem) (eqStateRz (memF M) a (rZComp b)) ->
  (eval M a)=(rZComp (eval M b)).
Intros a b M H'.
Rewrite <- evalrZComp; Auto.
Qed.
Hints Resolve evalCorrectComp.

Theorem evalInv:
  (a, b:rZ) (M:Mem) (eval M a)=(eval M b) ->(eqStateRz (memF M) a b).
Intros a b M H'.
Apply eqStateRzTrans with b := (eval M a); Auto.
Rewrite H'; Auto.
Qed.
Hints Immediate evalInv.

Theorem evalInvComp:
  (a, b:rZ) (M:Mem) (eval M a)=(rZComp (eval M b)) ->
  (eqStateRz (memF M) a (rZComp b)).
Intros a b M H'.
Apply evalInv; Auto.
Rewrite H'; Auto.
Qed.
Hints Immediate evalInvComp.

Theorem addEqCompGen:
  (a, b, a', b':rZ)
  (S, S':State) (eqStateRz S a a') -> (eqStateRz S b b') -> (eqState S S') ->
  (eqState (addEq (a, b) S) (addEq (a', b') S')).
Intros a b a' b' S S' H' H'0 H'1.
Apply eqStateTrans with S2 := (addEq (a', b') S); Auto.
Qed.
Hints Resolve addEqCompGen.

Theorem evalrZFalse: (M:Mem)(eval M rZFalse)=rZFalse.
Intros M; Change <rZ> (eval M (rZComp rZTrue))=rZFalse.
Rewrite evalrZComp; Auto.
Rewrite evalrZTrue; Auto.
Qed.

Theorem addCorrectExEq:
  (M:Mem) (a, b:rZ)(eqState (addEq (a, b) (memF M)) (memFx (add a b M))).
Intros M a b; Unfold memFx.
Generalize (refl_equal ? (add a b M)); Pattern -2 (add a b M); Case (add a b M);
 Simpl.
Intros L1; Case L1.
Intros M1 b1 L2.
Generalize (refl_equal ? b1); Pattern -2 b1; Case b1; Simpl.
Intros H' H'0.
Red; Split; Auto.
Red; Auto.
Intros i j H'1.
Apply eqStateContr with p := a q := b; Auto.
Unfold addEq; Apply eqStateRzTail; Auto.
Apply evalInvComp; Auto.
Apply addCorrectContrEval; Auto.
Rewrite <- H'0; Auto.
Intros H' H'0.
Generalize (addCorrectNotContrEq M a b).
Rewrite <- H'0; Auto.
Qed.
Hints Resolve addCorrectExEq.

Definition add2 :=
   [a, b, c, d:rZ] [M:Mem]
      Cases (add c d M) of
          (pair (pair M' b') L') => Cases b' of
                                        true => ((M', b'), L')
                                       | false => Cases (add a b M') of
                                                      (pair (pair M'' b'') L'')
                                                        =>
                                                        ((M'', b''),
                                                         (app L' L''))
                                                  end
                                    end
      end.

Theorem eqExceptOnTrans:
  (M1, M2, M3:Mem)
  (L1, L2:(list rZ)) (eqExceptOn M1 M2 L1) -> (eqExceptOn M2 M3 L2) ->
  (eqExceptOn M1 M3 (app L2 L1)).
Intros M1 M2 M3 L1 L2; Unfold eqExceptOn.
Intros H' H'0 c H'1 H'2.
Rewrite H'; Auto.
Apply H'0; Auto.
Red; Intros H'3; Case H'1; Auto with datatypes.
Red; Intros H'3; Case H'2; Auto with datatypes.
Red; Intros H'3; Case H'1; Auto with datatypes.
Red; Intros H'3; Case H'2; Auto with datatypes.
Qed.

Theorem addEq2SumEq:
  (S:State) (a, b, c, d:rZ)
  (eqState (addEq (a, b) (addEq (c, d) S)) (addEq (c, d) (addEq (a, b) S))).
Intros S a b c d.
Apply eqStateTrans
     with S2 :=
          (unionState
             (unionState (cons (c, d) (nil ?)) (cons (a, b) (nil ?))) S); Auto.
Apply eqStateTrans
     with S2 :=
          (unionState
             (unionState (cons (a, b) (nil ?)) (cons (c, d) (nil ?))) S); Auto.
Qed.
Hints Resolve addEq2SumEq.

Theorem add2In:
  (M:Mem) (p, q, r, s:rZ)
  (eqExceptOn (getMem (add2 p q r s M)) M (getList (add2 p q r s M))).
Intros M p q r s; Unfold add2; Auto.
Generalize (refl_equal ? (add r s M)); Pattern -1 (add r s M); Case (add r s M);
 Auto.
Intros Mb; Case Mb; Clear Mb.
Intros m b1; Case b1.
Unfold getMem; Simpl; Auto.
Intros l H'.
Generalize (addCorrectEval M r s); Rewrite H'; Auto.
Intros L1 H'.
Generalize (refl_equal ? (add p q m)); Pattern -1 (add p q m); Case (add p q m);
 Auto.
Intros Mb; Case Mb; Clear Mb.
Unfold getMem; Simpl; Auto.
Intros m0 b l H'0.
Apply eqExceptOnTrans with M2 := m; Auto.
Generalize (addCorrectEval m p q); Rewrite H'0; Unfold getMem; Simpl; Auto.
Generalize (addCorrectEval M r s); Rewrite H'; Unfold getMem; Simpl; Auto.
Qed.
Hints Resolve add2In.

Theorem ltWithListTrans:
  (M1, M2, M3:Mem)
  (L1, L2:(list rZ)) (ltWithList M1 M2 L1) -> (ltWithList M2 M3 L2) ->
  (ltWithList M1 M3 (app L2 L1)).
Intros M1 M2 M3 L1 L2; Case L1; Case L2; Simpl; Auto.
Intros H' H'0.
Apply eqMemTrans with y := M2; Auto.
Intros H' H'0 H'1 H'2.
Apply ltMemEqComp with M1 := M2 M3 := M3; Auto.
Apply eqMemSym; Auto.
Intros H' H'0 H'1 H'2.
Apply ltMemEqComp with M1 := M1 M3 := M2; Auto.
Intros H' H'0 H'1 H'2 H'3 H'4.
Apply lttrans with y := M2; Auto.
Qed.

Theorem add2ltWithList:
  (M:Mem) (p, q, r, s:rZ)
  (ltWithList (getMem (add2 p q r s M)) M (getList (add2 p q r s M))).
Intros M p q r s; Unfold add2.
Generalize (refl_equal ? (add r s M)); Pattern -1 (add r s M); Case (add r s M);
 Auto.
Intros Mb; Case Mb; Clear Mb.
Intros m b1; Case b1.
Unfold getMem; Simpl; Auto.
Intros l H'.
Generalize (addCorrectlt M r s); Rewrite H'; Simpl; Auto.
Intros L1 H'.
Generalize (refl_equal ? (add p q m)); Pattern -1 (add p q m); Case (add p q m);
 Auto.
Intros Mb; Case Mb; Clear Mb.
Unfold getMem; Simpl; Auto.
Intros m0 b l H'0.
Apply ltWithListTrans with M2 := m; Auto.
Generalize (addCorrectlt m p q); Rewrite H'0; Simpl; Auto with datatypes.
Generalize (addCorrectlt M r s); Rewrite H'; Simpl; Auto.
Qed.
Hints Resolve add2ltWithList.

Theorem doTripletFAux1:
  (M:Mem)
  (t:triplet) (p, q:rZ) (doTripletP (memF M) t (addEq (p, q) (memF M))) ->
  (ex
     State
     [S:State]
     (doTripletP (memF M) t S) /\
     (eqState S (memFx (add (eval M p) (eval M q) M)))).
Intros M t p q H'; Exists (addEq (p, q) (memF M)); Split; Auto.
Apply eqStateTrans with S2 := (addEq ((eval M p), (eval M q)) (memF M)); Auto.
Qed.
Hints Resolve doTripletFAux1.

Theorem doTripletFAux2:
  (M:Mem)
  (t:triplet)
  (p, q, r, s:rZ)
  (doTripletP (memF M) t (addEq (p, q) (addEq (r, s) (memF M)))) ->
  (ex
     State
     [S:State]
     (doTripletP (memF M) t S) /\
     (eqState S (memFx (add2 (eval M p) (eval M q) (eval M r) (eval M s) M)))).
Intros M t p q r s H'; Exists (addEq (p, q) (addEq (r, s) (memF M))); Split;
 Auto.
Apply eqStateTrans
     with S2 :=
          (addEq
             ((eval M p), (eval M q)) (addEq ((eval M r), (eval M s)) (memF M)));
 Auto.
Apply eqStateTrans
     with S2 :=
          (addEq ((eval M p), (eval M q)) (memFx (add (eval M r) (eval M s) M)));
 Auto.
Unfold add2.
Case (add (eval M r) (eval M s) M).
Simpl; Auto.
Intros Mb; Case Mb; Clear Mb.
Intros m b L1; Case b; Simpl; Auto.
Red; Split; Auto.
Generalize (addCorrectExEq m (eval M p) (eval M q)).
Case (add (eval M p) (eval M q) m); Simpl; Auto.
Intros Mb; Case Mb; Clear Mb; Simpl; Auto.
Qed.
Hints Resolve doTripletFAux2.

Definition doTripletF: (M:Mem) (t:triplet){R:(Option ?) | Cases R of
                                                              (Some R') =>
                                                                (Ex [S:State]
                                                                (doTripletP
                                                                   (memF M) t S)
                                                                /\
                                                                (eqState
                                                                   S (memFx R')))
                                                                /\
                                                                ((eqExceptOn
                                                                    (getMem R')
                                                                    M
                                                                    (getList R'))
                                                                 /\
                                                                 (ltWithList
                                                                    (getMem R')
                                                                    M
                                                                    (getList R')))
                                                             | None =>
                                                                 (S:State)
                                                                 ~ (doTripletP
                                                                      (memF M) t
                                                                      S)
                                                          end}.
Intros M t.
Case t; Intros b p q r.
Apply LetP with h := (eval M p); Intros p' Hp'.
Apply LetP with h := (eval M q); Intros q' Hq'.
Apply LetP with h := (eval M r); Intros r' Hr'.
Case b.
Case (rZDec p' (rZComp q')); Rewrite Hp'; Rewrite Hq'; Intros Eq1.
Exists (Some ? (add2 q' rZTrue r' rZFalse M)).
Rewrite Hq'; Rewrite Hr'; Rewrite <- (evalrZTrue M); Rewrite <- (evalrZFalse M);
 Auto.
Case (rZDec p' (rZComp r')); Rewrite Hp'; Rewrite Hr'; Intros Eq2; Repeat Split;
 Auto.
Exists (Some ? (add2 q' rZFalse r' rZTrue M)).
Rewrite Hq'; Rewrite Hr'; Rewrite <- (evalrZTrue M); Rewrite <- (evalrZFalse M);
 Auto.
Case (rZDec q' r'); Rewrite Hq'; Rewrite Hr'; Intros Eq3.
Exists (Some ? (add p' r' M)).
Rewrite Hp'; Rewrite Hr'; Auto.
Case (rZDec q' (rZComp r')); Rewrite Hq'; Rewrite Hr'; Intros Eq4.
Exists (Some ? (add p' rZFalse M)).
Rewrite Hp'; Rewrite <- (evalrZFalse M); Auto.
Case (rZDec p' rZTrue); Rewrite Hp'; Rewrite <- (evalrZTrue M); Intros Eq5.
Exists (Some ? (add2 q' rZTrue r' rZTrue M)).
Rewrite Hq'; Rewrite Hr'; Rewrite <- (evalrZTrue M); Auto.
Case (rZDec q' rZTrue); Rewrite Hq'; Rewrite <- (evalrZTrue M); Intros Eq6.
Exists (Some ? (add p' r' M)).
Rewrite Hp'; Rewrite Hr'; Auto.
Case (rZDec q' rZFalse); Rewrite Hq'; Rewrite <- (evalrZFalse M); Intros Eq7.
Exists (Some ? (add p' rZFalse M)).
Rewrite Hp'; Rewrite <- (evalrZFalse M); Auto.
Case (rZDec r' rZTrue); Rewrite Hr'; Rewrite <- (evalrZTrue M); Intros Eq8.
Exists (Some ? (add p' q' M)).
Rewrite Hp'; Rewrite Hq'; Auto.
Case (rZDec r' rZFalse); Rewrite Hr'; Rewrite <- (evalrZFalse M); Intros Eq9.
Exists (Some ? (add p' rZFalse M)).
Rewrite Hp'; Rewrite <- (evalrZFalse M); Auto.
Exists (None (Mem * bool) * (list rZ)).
Intros S; Red; Intros H'; Inversion H'; Auto.
Case (rZDec p' q'); Rewrite Hp'; Rewrite Hq'; Intros Eq1.
Exists (Some ? (add r' rZTrue M)).
Rewrite Hr'; Rewrite <- (evalrZTrue M); Auto.
Case (rZDec p' (rZComp q')); Rewrite Hp'; Rewrite Hq'; Intros Eq2.
Exists (Some ? (add r' rZFalse M)).
Rewrite Hr'; Rewrite <- (evalrZFalse M); Auto.
Case (rZDec p' r'); Rewrite Hp'; Rewrite Hr'; Intros Eq3.
Exists (Some ? (add q' rZTrue M)).
Rewrite Hq'; Rewrite <- (evalrZTrue M); Auto.
Case (rZDec p' (rZComp r')); Rewrite Hp'; Rewrite Hr'; Intros Eq4.
Exists (Some ? (add q' rZFalse M)).
Rewrite Hq'; Rewrite <- (evalrZFalse M); Auto.
Case (rZDec q' r'); Rewrite Hq'; Rewrite Hr'; Intros Eq5.
Exists (Some ? (add p' rZTrue M)).
Rewrite Hp'; Rewrite <- (evalrZTrue M); Auto.
Case (rZDec q' (rZComp r')); Rewrite Hq'; Rewrite Hr'; Intros Eq6.
Exists (Some ? (add p' rZFalse M)).
Rewrite Hp'; Rewrite <- (evalrZFalse M); Auto.
Case (rZDec p' rZTrue); Rewrite Hp'; Rewrite <- (evalrZTrue M); Intros Eq7.
Exists (Some ? (add q' r' M)).
Rewrite Hq'; Rewrite Hr'; Auto.
Case (rZDec p' rZFalse); Rewrite Hp'; Rewrite <- (evalrZFalse M); Intros Eq8.
Exists (Some ? (add q' (rZComp r') M)).
Rewrite Hq'; Rewrite Hr'; Rewrite <- evalrZComp; Auto.
Case (rZDec q' rZTrue); Rewrite Hq'; Rewrite <- (evalrZTrue M); Intros Eq9.
Exists (Some ? (add p' r' M)).
Rewrite Hp'; Rewrite Hr'; Auto.
Case (rZDec q' rZFalse); Rewrite Hq'; Rewrite <- (evalrZFalse M); Intros Eq10.
Exists (Some ? (add p' (rZComp r') M)).
Rewrite Hp'; Rewrite Hr'; Rewrite <- evalrZComp; Auto.
Case (rZDec r' rZTrue); Rewrite Hr'; Rewrite <- (evalrZTrue M); Intros Eq11.
Exists (Some ? (add p' q' M)).
Rewrite Hp'; Rewrite Hq'; Auto.
Case (rZDec r' rZFalse); Rewrite Hr'; Rewrite <- (evalrZFalse M); Intros Eq12.
Exists (Some ? (add p' (rZComp q') M)).
Rewrite Hp'; Rewrite Hq'; Rewrite <- evalrZComp; Auto.
Exists (None (Mem * bool) * (list rZ)).
Intros S; Red; Intros H'; Inversion H'; Auto.
Defined.

Definition doTripletsAuxF:
  (L:(list triplet)) (M:Mem)
  {R:(Mem * bool) * (list rZ) |
    (doTripletsP (memF M) L (memFx R)) /\
    ((eqExceptOn (getMem R) M (getList R)) /\
     (ltWithList (getMem R) M (getList R)))}.
Intros L1; Elim L1; Clear L1.
Intros M; Exists ((M, false), (nil rZ)); Repeat Split; Auto.
Intros t L2 Rec M.
Case (doTripletF M t).
Intros x; Case x; Clear x.
Intros R1; Generalize (refl_equal ? R1); Pattern -1 R1; Case R1.
Intros x; Case x; Clear x.
Intros M1 b1 L1; Case b1.
Intros EqR1 H'1; Elim H'1; Intros H'2 H'3; Elim H'3; Intros H'4 H'5;
 Clear H'3 H'1; Auto.
Exists R1; Rewrite EqR1; Repeat Split; Auto.
Elim H'2; Intros S E; Elim E; Intros H'0 H'1; Clear E H'2.
Apply doTripletsTrans with S2 := S t := t; Auto with datatypes.
Intros EqR1 H'; Case (Rec M1).
Intros R2; Generalize (refl_equal ? R2); Pattern -1 R2; Case R2.
Intros Mb L3 EqL2 H'0; Exists (Mb, (app L1 L3)); Auto.
Elim H'; Intros H'1 H'2; Elim H'2; Intros H'3 H'4; Clear H'2 H'.
Elim H'0; Intros H' H'2; Elim H'2; Intros H'5 H'6; Clear H'2 H'0.
Repeat Split; Auto.
Elim H'1; Intros S E; Elim E; Intros H'0 H'2; Clear E H'1.
Apply doTripletsTrans with S2 := S t := t; Auto with datatypes.
Apply doTripletsComp
     with S1 := (memF M1) S2 := (memFx <Mem * bool, (list rZ)> (Mb, L3)); Auto.
Apply doTripletsInclList with L1 := L2; Auto with datatypes.
Unfold getList getMem; Simpl.
Apply eqExceptOnTrans with M2 := M1; Auto.
Unfold getList getMem; Simpl.
Apply ltWithListTrans with M2 := M1; Auto.
Intros H'; Case (Rec M).
Intros Res H'0; Elim H'0; Intros H'1 H'2; Elim H'2; Intros H'3 H'4;
 Clear H'2 H'0.
Exists Res; Repeat Split; Auto.
Apply doTripletsInclList with L1 := L2; Auto with datatypes.
Qed.

Definition doTripletsF:
  (M:Mem) (Lv:(list rZ))
  {R:(Mem * bool) * (list rZ) |
    (doTripletsP (memF M) L (memFx R)) /\
    ((eqExceptOn (getMem R) M (getList R)) /\
     (ltWithList (getMem R) M (getList R)))}.
Intros M; Pattern M; Apply well_founded_induction with A := Mem R := ltMem;
 Clear M; Auto.
Intros M RecLt L1; Elim L1; Clear L1.
Exists ((M, false), (nil rZ)); Repeat Split; Auto.
Intros v Lv RecLv.
Case (doTripletsAuxF (varList v) M).
Intros R1; Generalize (refl_equal ? R1); Pattern -1 R1; Case R1.
Intros x; Case x; Clear x.
Intros M1 b1 L1; Case b1.
Intros H' H'0; Elim H'0; Intros H'1 H'2; Elim H'2; Intros H'3 H'4;
 Clear H'2 H'0; Exists R1; Repeat Split; Try Rewrite H'; Simpl; Auto.
Apply doTripletsInclList with L1 := (varList v); Auto with datatypes.
Red; Intros a H'1.
Apply varListCorrectIn with a := v; Auto.
Generalize (refl_equal ? L1); Pattern -1 L1; Case L1.
Intros H' H'0 H'1; Exact RecLv.
Intros r l H' H'0 H'1; Elim H'1; Intros H'2 H'3; Elim H'3; Intros H'4 H'5;
 Clear H'3 H'1.
Cut (ltMem M1 M); [Intros ltM1 | Idtac].
Case (RecLt M1 ltM1 (app L1 Lv)).
Intros x; Case x; Clear x.
Intros Mb2 L2 H'1; Elim H'1; Intros H'3 H'6; Elim H'6; Intros H'7 H'8;
 Clear H'6 H'1.
Exists (Mb2, (app L1 L2)); Rewrite H'; Repeat Split; Auto.
Apply doTripletsRTrans with S2 := (memF M1); Auto.
Apply doTripletsInclList with L1 := (varList v); Auto.
Red; Auto.
Intros a H'2.
Apply varListCorrectIn with a := v; Auto.
Change (eqExceptOn (Fst Mb2) M (app (cons r l) L2)).
Apply eqExceptOnTrans with M2 := M1; Auto.
Change (ltWithList (Fst Mb2) M (app (cons r l) L2)).
Apply ltWithListTrans with M2 := M1; Auto.
Unfold ltWithList getMem in H'5; Simpl in H'5; Auto.
Qed.

Theorem eqqMemImpEqState:
  (M1, M2:Mem) (eqMem M1 M2) ->(eqState (memF M1) (memF M2)).
Intros M1 M2 H'; Red in H'; Red; Split; Red; Intros i j H; Apply evalInv; Auto.
Repeat Rewrite <- H'; Apply evalCorrect; Auto.
Repeat Rewrite H'; Apply evalCorrect; Auto.
Qed.

Definition dilemmaF:
  Mem -> Mem -> Mem -> rZ ->((Mem * (Mem * Mem)) * bool) * bool :=
   [M1, M2, M3:Mem] [a:rZ]
      Cases (add a rZTrue M2) of
          (pair (pair M'1 b'1) L'1) =>
            (Case b'1 of
              (((M1, ((reset M'1 M1 L'1), M3)), false), false)
              Cases (add a rZFalse M3) of
                  (pair (pair M'2 b'2) L'2) => Cases b'2 of
                                                   true =>
                                                     (((M1,
                                                        (M2, (reset M'2 M1 L'2))),
                                                       false), false)
                                                  | false =>
                                                      Cases (doTripletsF
                                                               M'1 L'1) of
                                                          (exist
                                                             (pair
                                                                (pair M'3 b'3)
                                                                  L'3) _) =>
                                                            Cases (doTripletsF
                                                                     M'2 L'2) of
                                                                (exist
                                                                   (pair
                                                                      (pair
                                                                         M'4 b'4)
                                                                        L'4) _)
                                                                  =>
                                                                  (LetP
                                                                     ? ?
                                                                     (app
                                                                        L'1 L'3)
                                                                     [L'1L'3:?]
                                                                     [_:?]
                                                                     (LetP
                                                                        ? ?
                                                                        (app
                                                                           L'2
                                                                           L'4)
                                                                        [L'2L'4:?]
                                                                        [_:?]
                                                                           Cases b'3
                                                                                 b'4 of
                                                                               false
                                                                                 false
                                                                                 =>
                                                                                 Cases (inter
                                                                                          M1
                                                                                          M'3
                                                                                          M'4
                                                                                          L'1L'3
                                                                                          L'2L'4) of
                                                                                     (pair
                                                                                        M'5
                                                                                          L'5)
                                                                                       =>
                                                                                       (((M'5,
                                                                                          ((reset
                                                                                              M'3
                                                                                              M'5
                                                                                              L'1L'3),
                                                                                           (reset
                                                                                              M'4
                                                                                              M'5
                                                                                              L'2L'4))),
                                                                                         false),
                                                                                        Cases L'5 of
                                                                                            nil
                                                                                              =>
                                                                                              false
                                                                                           | _
                                                                                               =>
                                                                                               true
                                                                                        end)
                                                                                 end
                                                                              | true
                                                                                  false
                                                                                  =>
                                                                                  ((((reset
                                                                                        M1
                                                                                        M'4
                                                                                        L'2L'4),
                                                                                     ((reset
                                                                                         (reset
                                                                                            M'3
                                                                                            M'4
                                                                                            L'1L'3)
                                                                                         M'4
                                                                                         L'2L'4),
                                                                                      M'4)),
                                                                                    false),
                                                                                   Cases L'2L'4 of
                                                                                       nil
                                                                                         =>
                                                                                         false
                                                                                      | _
                                                                                          =>
                                                                                          true
                                                                                   end)
                                                                              | false
                                                                                  true
                                                                                  =>
                                                                                  ((((reset
                                                                                        M1
                                                                                        M'3
                                                                                        L'1L'3),
                                                                                     (M'3,
                                                                                      (reset
                                                                                         (reset
                                                                                            M'4
                                                                                            M'3
                                                                                            L'2L'4)
                                                                                         M'3
                                                                                         L'1L'3))),
                                                                                    false),
                                                                                   Cases L'1L'3 of
                                                                                       nil
                                                                                         =>
                                                                                         false
                                                                                      | _
                                                                                          =>
                                                                                          true
                                                                                   end)
                                                                              | true
                                                                                  true
                                                                                  =>
                                                                                  (((M1,
                                                                                     (M'3,
                                                                                      M'4)),
                                                                                    true),
                                                                                   false)
                                                                           end))
                                                            end
                                                      end
                                               end
              end end)
      end.

Theorem eqExceptEq:
  (M1, M2:Mem) (L:(list rZ)) (eqExceptOn M1 M2 L) ->(eqMem (reset M1 M2 L) M2).
Unfold eqExceptOn; Intros M1 M2 L0 H'; Red; Intros c.
Case (In_dec rZDec c L0); Intros IncL0; Auto.
Case (In_dec rZDec (rZComp c) L0); Intros InmcL0; Auto.
Rewrite <- H'; Auto.
Qed.

Theorem dilemmaFProp1:
  (M1, M2, M3:Mem) (eqMem M1 M2) -> (eqMem M1 M3) ->
  (a:rZ)Cases (dilemmaF M1 M2 M3 a) of
            (pair (pair (pair M'1 (pair M'2 M'3)) b1) true) => (ltMem M'1 M1)
           | (pair (pair (pair M'1 (pair M'2 M'3)) b1) false) => (eqMem M'1 M1)
        end.
Intros M1 M2 M3 H' H'0 a; Unfold dilemmaF.
Generalize (refl_equal ? (add a rZTrue M2)); Pattern -1 (add a rZTrue M2);
 Case (add a rZTrue M2).
Intros Mb L'1; Case Mb; Clear Mb; Intros M'1 b'1; Case b'1; Intros EqM'1; Simpl;
 Auto.
Generalize (refl_equal ? (add a rZFalse M3)); Pattern -1 (add a rZFalse M3);
 Case (add a rZFalse M3).
Intros Mb L'2; Case Mb; Clear Mb; Intros M'2 b'2; Case b'2; Intros EqM'2; Simpl;
 Auto.
Case (doTripletsF M'1 L'1).
Intros x; Case x; Clear x.
Intros x; Case x; Clear x.
Intros M'3 b'3 L'3 H'.
Case (doTripletsF M'2 L'2).
Intros x; Case x; Clear x.
Intros x; Case x; Clear x.
Intros M'4 b'4 L'4 H'1.
Unfold LetP; Auto.
Case b'3; Case b'4; Auto.
Cut (ltWithList M'4 M1 (app L'2 L'4)).
Cut (eqMem (reset M1 M'4 (app L'2 L'4)) M'4).
Case (app L'2 L'4); Unfold ltWithList; Auto.
Intros H'3 H'4; Red; Auto.
Intros r l H'3 H'4.
Apply ltMemEqComp with M1 := M'4 M3 := M1; Auto.
Apply eqMemSym; Auto.
Apply eqExceptEq; Auto.
Rewrite appSym; Auto.
Apply eqExceptOnTrans with M2 := M'2; Auto.
Change (eqExceptOn M1 M'2 (app (nil ?) L'2)).
Rewrite appSym; Auto.
Apply eqExceptOnTrans with M2 := M3; Auto.
Generalize (addCorrectEval M3 a rZFalse).
Rewrite EqM'2; Auto.
Elim H'2; Intros H'3 H'4; Elim H'4; Intros H'5 H'6; Clear H'4 H'1; Auto.
Apply ltWithListTrans with M2 := M'2; Auto.
Elim H'2; Intros H'3 H'4; Elim H'4; Intros H'5 H'6; Clear H'4 H'1; Auto.
Change (ltWithList M'2 M1 (app (nil ?) L'2)).
Apply ltWithListTrans with M2 := M3; Auto.
Generalize (addCorrectlt M3 a rZFalse).
Rewrite EqM'2; Auto.
Red; Simpl; Auto.
Apply eqMemSym; Auto.
Cut (ltWithList M'3 M1 (app L'1 L'3)).
Cut (eqMem (reset M1 M'3 (app L'1 L'3)) M'3).
Case (app L'1 L'3); Unfold ltWithList; Auto.
Intros H'3 H'4; Red; Auto.
Intros r l H'3 H'4.
Apply ltMemEqComp with M1 := M'3 M3 := M1; Auto.
Apply eqMemSym; Auto.
Apply eqExceptEq; Auto.
Rewrite appSym; Auto.
Apply eqExceptOnTrans with M2 := M'1; Auto.
Change (eqExceptOn M1 M'1 (app (nil ?) L'1)).
Apply eqExceptOnSym.
Apply eqExceptOnTrans with M2 := M2; Auto.
Generalize (addCorrectEval M2 a rZTrue).
Rewrite EqM'1; Auto.
Apply eqExceptOnSym; Auto.
Elim H'1; Intros H'3 H'4; Elim H'4; Intros H'5 H'6; Clear H'4 H'1; Auto.
Apply ltWithListTrans with M2 := M'1; Auto.
Elim H'1; Intros H'3 H'4; Elim H'4; Intros H'5 H'6; Clear H'4 H'1; Auto.
Change (ltWithList M'1 M1 (app (nil ?) L'1)).
Apply ltWithListTrans with M2 := M2; Auto.
Generalize (addCorrectlt M2 a rZTrue).
Rewrite EqM'1; Auto.
Red; Simpl; Auto.
Apply eqMemSym; Auto.
Generalize (refl_equal ? (inter M1 M'3 M'4 (app L'1 L'3) (app L'2 L'4)));
 Pattern -1 (inter M1 M'3 M'4 (app L'1 L'3) (app L'2 L'4));
 Case (inter M1 M'3 M'4 (app L'1 L'3) (app L'2 L'4)).
Intros M'5 L'5; Case L'5; [Idtac | Intros a6 L6]; Intros H'3;
 Generalize (interlt M1 M'3 M'4 (app L'1 L'3) (app L'2 L'4)); Auto; Rewrite H'3;
 Auto.
Qed.

Theorem dilemmaFProp2:
  (M1, M2, M3:Mem) (eqMem M1 M2) -> (eqMem M1 M3) ->
  (a:rZ)Cases (dilemmaF M1 M2 M3 a) of
            (pair (pair (pair M'1 (pair M'2 M'3)) true) b2) =>
              (dilemmaP
                 (memF M1) L
                 (cons ((rZPlus zero), (rZMinus zero)) (nil rZ * rZ)))
           | (pair (pair (pair M'1 (pair M'2 M'3)) false) b2) =>
               (dilemmaP (memF M1) L (memF M'1))
        end.
Intros M1 M2 M3 H' H'0 a; Unfold dilemmaF.
Generalize (refl_equal ? (add a rZTrue M2)); Pattern -1 (add a rZTrue M2);
 Case (add a rZTrue M2).
Intros Mb L'1; Case Mb; Clear Mb; Intros M'1 b'1; Case b'1; Intros EqM'1; Simpl;
 Auto.
Generalize (refl_equal ? (add a rZFalse M3)); Pattern -1 (add a rZFalse M3);
 Case (add a rZFalse M3).
Intros Mb L'2; Case Mb; Clear Mb; Intros M'2 b'2; Case b'2; Intros EqM'2; Simpl;
 Auto.
Case (doTripletsF M'1 L'1).
Intros x; Case x; Clear x.
Intros x; Case x; Clear x.
Intros M'3 b'3 L'3 H'.
Elim H'1; Intros H'2 H'3; Elim H'3; Intros H'4 H'5; Clear H'3 H'1.
Case (doTripletsF M'2 L'2).
Intros x; Case x; Clear x.
Intros x; Case x; Clear x.
Intros M'4 b'4 L'4 H'1.
Elim H'1; Intros H'3 H'6; Elim H'6; Intros H'7 H'8; Clear H'6 H'1.
Unfold LetP; Auto.
Cut (eqExceptOn M'3 M1 (app L'1 L'3)); [Intros eqEx1 | Idtac].
Cut (eqExceptOn M'4 M1 (app L'2 L'4)); [Intros eqEx2 | Idtac].
Cut (ltWithList M'4 M1 (app L'2 L'4)); [Intros ltM2 | Idtac].
Cut (eqMem (reset M1 M'4 (app L'2 L'4)) M'4); [Intros eqM2 | Idtac].
Cut (ltWithList M'3 M1 (app L'1 L'3)); [Intros ltM1 | Idtac].
Cut (eqMem (reset M1 M'3 (app L'1 L'3)) M'3); [Intros eqM1 | Idtac].
Generalize H'3 H'2; Case b'3; Case b'4; Clear H'3 H'2; Simpl; Auto;
 Intros H'3 H'2.
Apply dilemmaPSplit
     with a := a b := rZTrue
          S2 := (cons ((rZPlus zero), (rZMinus zero)) (nil ?))
          S3 := (cons ((rZPlus zero), (rZMinus zero)) (nil ?)); Auto.
Apply dilemmaPRef; Auto.
Apply doTripletsComp
     with S1 := (memFx (add a rZTrue M2))
          S2 := (cons ((rZPlus zero), (rZMinus zero)) (nil ?)); Auto.
Rewrite EqM'1; Auto.
Apply eqStateTrans with S2 := (addEq (a, rZTrue) (memF M2)); Auto.
Apply eqStateAddEq; Auto.
Apply eqqMemImpEqState; Auto.
Apply dilemmaPRef; Auto.
Apply doTripletsComp
     with S1 := (memFx (add a rZFalse M3))
          S2 := (cons ((rZPlus zero), (rZMinus zero)) (nil ?)); Auto.
Rewrite EqM'2; Auto.
Apply eqStateTrans with S2 := (addEq (a, rZFalse) (memF M3)); Auto.
Apply eqStateAddEq; Auto.
Apply eqqMemImpEqState; Auto.
Apply dilemmaPSplit
     with a := a b := rZTrue
          S2 := (cons ((rZPlus zero), (rZMinus zero)) (nil ?)) S3 := (memF M'4);
 Auto.
Apply dilemmaPRef; Auto.
Apply doTripletsComp
     with S1 := (memFx (add a rZTrue M2))
          S2 := (cons ((rZPlus zero), (rZMinus zero)) (nil ?)); Auto.
Rewrite EqM'1; Auto.
Apply eqStateTrans with S2 := (addEq (a, rZTrue) (memF M2)); Auto.
Apply eqStateAddEq; Auto.
Apply eqqMemImpEqState; Auto.
Apply dilemmaPRef; Auto.
Apply doTripletsComp with S1 := (memFx (add a rZFalse M3)) S2 := (memF M'4);
 Auto.
Rewrite EqM'2; Auto.
Apply eqStateTrans with S2 := (addEq (a, rZFalse) (memF M3)); Auto.
Apply eqStateAddEq; Auto.
Apply eqqMemImpEqState; Auto.
Apply eqStateTrans with S2 := (memF M'4); Auto.
Apply eqqMemImpEqState; Auto.
Apply eqMemSym; Auto.
Apply dilemmaPSplit
     with a := a b := rZTrue S2 := (memF M'3)
          S3 := (cons ((rZPlus zero), (rZMinus zero)) (nil ?)); Auto.
Apply dilemmaPRef; Auto.
Apply doTripletsComp with S1 := (memFx (add a rZTrue M2)) S2 := (memF M'3); Auto.
Rewrite EqM'1; Auto.
Apply eqStateTrans with S2 := (addEq (a, rZTrue) (memF M2)); Auto.
Apply eqStateAddEq; Auto.
Apply eqqMemImpEqState; Auto.
Apply dilemmaPRef; Auto.
Apply doTripletsComp
     with S1 := (memFx (add a rZFalse M3))
          S2 := (cons ((rZPlus zero), (rZMinus zero)) (nil ?)); Auto.
Rewrite EqM'2; Auto.
Apply eqStateTrans with S2 := (addEq (a, rZFalse) (memF M3)); Auto.
Apply eqStateAddEq; Auto.
Apply eqqMemImpEqState; Auto.
Apply eqStateTrans with S2 := (memF M'3); Auto.
Apply eqqMemImpEqState; Auto.
Apply eqMemSym; Auto.
Generalize (refl_equal ? (inter M1 M'3 M'4 (app L'1 L'3) (app L'2 L'4)));
 Pattern -1 (inter M1 M'3 M'4 (app L'1 L'3) (app L'2 L'4));
 Case (inter M1 M'3 M'4 (app L'1 L'3) (app L'2 L'4)).
Intros M'5 L'5; Case L'5; Auto.
Intros EqIn.
Apply dilemmaPSplit with a := a b := rZTrue S2 := (memF M'3) S3 := (memF M'4);
 Auto.
Apply dilemmaPRef; Auto.
Apply doTripletsComp with S1 := (memFx (add a rZTrue M2)) S2 := (memF M'3); Auto.
Rewrite EqM'1; Auto.
Apply eqStateTrans with S2 := (addEq (a, rZTrue) (memF M2)); Auto.
Apply eqStateAddEq; Auto.
Apply eqqMemImpEqState; Auto.
Apply dilemmaPRef; Auto.
Apply doTripletsComp with S1 := (memFx (add a rZFalse M3)) S2 := (memF M'4);
 Auto.
Rewrite EqM'2; Auto.
Apply eqStateTrans with S2 := (addEq (a, rZFalse) (memF M3)); Auto.
Apply eqStateAddEq; Auto.
Apply eqqMemImpEqState; Auto.
Apply eqStateSym.
Apply eqStateTrans
     with S2 := (memF (Fst (inter M1 M'3 M'4 (app L'1 L'3) (app L'2 L'4))));
 Auto.
Rewrite EqIn; Auto.
Intros a'6 L'6 EqIn.
Apply dilemmaPSplit with a := a b := rZTrue S2 := (memF M'3) S3 := (memF M'4);
 Auto.
Apply dilemmaPRef; Auto.
Apply doTripletsComp with S1 := (memFx (add a rZTrue M2)) S2 := (memF M'3); Auto.
Rewrite EqM'1; Auto.
Apply eqStateTrans with S2 := (addEq (a, rZTrue) (memF M2)); Auto.
Apply eqStateAddEq; Auto.
Apply eqqMemImpEqState; Auto.
Apply dilemmaPRef; Auto.
Apply doTripletsComp with S1 := (memFx (add a rZFalse M3)) S2 := (memF M'4);
 Auto.
Rewrite EqM'2; Auto.
Apply eqStateTrans with S2 := (addEq (a, rZFalse) (memF M3)); Auto.
Apply eqStateAddEq; Auto.
Apply eqqMemImpEqState; Auto.
Apply eqStateSym.
Apply eqStateTrans
     with S2 := (memF (Fst (inter M1 M'3 M'4 (app L'1 L'3) (app L'2 L'4))));
 Auto.
Rewrite EqIn; Auto.
Apply eqExceptEq; Auto.
Apply ltWithListTrans with M2 := M'1; Auto.
Change (ltWithList M'1 M1 (app (nil ?) L'1)).
Apply ltWithListTrans with M2 := M2; Auto.
Generalize (addCorrectlt M2 a rZTrue); Rewrite EqM'1; Auto.
Red; Auto.
Apply eqMemSym; Auto.
Apply eqExceptEq; Auto.
Apply ltWithListTrans with M2 := M'2; Auto.
Change (ltWithList M'2 M1 (app (nil ?) L'2)).
Apply ltWithListTrans with M2 := M3; Auto.
Generalize (addCorrectlt M3 a rZFalse); Rewrite EqM'2; Auto.
Red; Auto.
Apply eqMemSym; Auto.
Apply eqExceptOnTrans with M2 := M'2; Auto.
Change (eqExceptOn M'2 M1 (app (nil ?) L'2)).
Apply eqExceptOnTrans with M2 := M3; Auto.
Generalize (addCorrectEval M3 a rZFalse); Rewrite EqM'2; Auto.
Apply eqExceptOnRef; Auto.
Apply eqMemSym; Auto.
Apply eqExceptOnTrans with M2 := M'1; Auto.
Change (eqExceptOn M'1 M1 (app (nil ?) L'1)).
Apply eqExceptOnTrans with M2 := M2; Auto.
Generalize (addCorrectEval M2 a rZTrue); Rewrite EqM'1; Auto.
Apply eqExceptOnRef; Auto.
Apply eqMemSym; Auto.
Qed.

Theorem dilemmaFProp3:
  (M1, M2, M3:Mem) (eqMem M1 M2) -> (eqMem M1 M3) ->
  (a:rZ)Cases (dilemmaF M1 M2 M3 a) of
            (pair (pair (pair M'1 (pair M'2 M'3)) false) b2) =>
              (eqMem M'1 M'2) /\ (eqMem M'1 M'3)
           | _ => True
        end.
Intros M1 M2 M3 H' H'0 a; Unfold dilemmaF.
Cut (eqExceptOn (getMem (add a rZTrue M2)) M1 (getList (add a rZTrue M2))).
Cut (eqExceptOn (getMem (add a rZFalse M3)) M1 (getList (add a rZFalse M3))).
Generalize (refl_equal ? (add a rZTrue M2)); Pattern -1 (add a rZTrue M2);
 Case (add a rZTrue M2).
Intros Mb L'1; Case Mb; Clear Mb; Intros M'1 b'1; Case b'1; Intros EqM'1;
 Unfold getMem; Simpl; Auto.
Intros H'1 H'2; Split; Auto.
Apply eqMemSym.
Apply eqExceptEq; Auto.
Generalize (refl_equal ? (add a rZFalse M3)); Pattern -1 (add a rZFalse M3);
 Case (add a rZFalse M3).
Intros Mb L'2; Case Mb; Clear Mb; Intros M'2 b'2; Case b'2; Intros EqM'2;
 Unfold getMem; Simpl; Auto.
Intros H'1 H'2; Split; Auto.
Apply eqMemSym.
Apply eqExceptEq; Auto.
Intros eqEx1 eqEx2.
Case (doTripletsF M'1 L'1).
Intros x; Case x; Clear x.
Intros x; Case x; Clear x.
Intros M'3 b'3 L'3 H'.
Elim H'1; Intros H'2 H'3; Elim H'3; Intros H'4 H'5; Clear H'3 H'1.
Case (doTripletsF M'2 L'2).
Intros x; Case x; Clear x.
Intros x; Case x; Clear x.
Intros M'4 b'4 L'4 H'1.
Elim H'1; Intros H'3 H'6; Elim H'6; Intros H'7 H'8; Clear H'6 H'1.
Unfold LetP; Auto.
Cut (eqExceptOn M1 M'3 (app L'1 L'3)); [Intros eqEx1 | Idtac].
Cut (eqExceptOn M1 M'4 (app L'2 L'4)); [Intros eqEx2 | Idtac].
Case b'3; Case b'4; Simpl; Try Split; Auto.
Apply eqMemTrans with y := M'4; Auto.
Apply eqExceptEq; Auto.
Red; Intros c; Case (In_dec rZDec c L'2); Intros IncL'2.
Rewrite resetClearIn; Auto with datatypes.
Case (In_dec rZDec (rZComp c) L'2); Intros InmcL'2.
Rewrite resetClearIn; Auto with datatypes.
Case (In_dec rZDec c L'4); Intros IncL'4.
Rewrite resetClearIn; Auto with datatypes.
Case (In_dec rZDec (rZComp c) L'4); Intros InmcL'4.
Rewrite resetClearIn; Auto with datatypes.
Rewrite resetClearNIn; Auto with datatypes.
Case (In_dec rZDec c L'1); Intros IncL'1.
Rewrite resetClearIn; Auto with datatypes.
Case (In_dec rZDec (rZComp c) L'1); Intros InmcL'1.
Rewrite resetClearIn; Auto with datatypes.
Case (In_dec rZDec c L'3); Intros IncL'3.
Rewrite resetClearIn; Auto with datatypes.
Case (In_dec rZDec (rZComp c) L'3); Intros InmcL'3.
Rewrite resetClearIn; Auto with datatypes.
Rewrite resetClearNIn; Auto with datatypes.
Apply trans_equal with y := (eval M1 c).
Apply sym_equal; Apply eqEx3; Auto; Red; Intros H'1; Case (in_app_or H'1); Auto.
Apply eqEx0; Auto; Red; Intros H'1; Case (in_app_or H'1); Auto.
Red; Intros H'1; Case (in_app_or H'1); Auto.
Red; Intros H'1; Case (in_app_or H'1); Auto.
Red; Intros H'1; Case (in_app_or H'1); Auto.
Red; Intros H'1; Case (in_app_or H'1); Auto.
Apply eqExceptEq; Auto.
Apply eqExceptEq; Auto.
Apply eqMemTrans with y := M'3; Auto.
Apply eqExceptEq; Auto.
Red; Intros c; Case (In_dec rZDec c L'1); Intros IncL'1.
Rewrite resetClearIn; Auto with datatypes.
Case (In_dec rZDec (rZComp c) L'1); Intros InmcL'1.
Rewrite resetClearIn; Auto with datatypes.
Case (In_dec rZDec c L'3); Intros IncL'3.
Rewrite resetClearIn; Auto with datatypes.
Case (In_dec rZDec (rZComp c) L'3); Intros InmcL'3.
Rewrite resetClearIn; Auto with datatypes.
Rewrite resetClearNIn; Auto with datatypes.
Case (In_dec rZDec c L'2); Intros IncL'2.
Rewrite resetClearIn; Auto with datatypes.
Case (In_dec rZDec (rZComp c) L'2); Intros InmcL'2.
Rewrite resetClearIn; Auto with datatypes.
Case (In_dec rZDec c L'4); Intros IncL'4.
Rewrite resetClearIn; Auto with datatypes.
Case (In_dec rZDec (rZComp c) L'4); Intros InmcL'4.
Rewrite resetClearIn; Auto with datatypes.
Rewrite resetClearNIn; Auto with datatypes.
Apply trans_equal with y := (eval M1 c).
Apply sym_equal; Apply eqEx0; Auto; Red; Intros H'1; Case (in_app_or H'1); Auto.
Apply eqEx3; Auto; Red; Intros H'1; Case (in_app_or H'1); Auto.
Red; Intros H'1; Case (in_app_or H'1); Auto.
Red; Intros H'1; Case (in_app_or H'1); Auto.
Red; Intros H'1; Case (in_app_or H'1); Auto.
Red; Intros H'1; Case (in_app_or H'1); Auto.
Generalize (refl_equal ? (inter M1 M'3 M'4 (app L'1 L'3) (app L'2 L'4)));
 Pattern -1 (inter M1 M'3 M'4 (app L'1 L'3) (app L'2 L'4));
 Case (inter M1 M'3 M'4 (app L'1 L'3) (app L'2 L'4)).
Intros M'5 L'5 EqM'5; Split; Auto.
Apply eqMemSym.
Apply eqExceptEq; Auto.
Apply eqExceptOnrTrans with M2 := M1; Auto.
Apply eqExceptOnIncl with L1 := L'5.
Generalize (interCorrectEq M1 M'3 M'4 (app L'1 L'3) (app L'2 L'4));
 Rewrite EqM'5; Simpl; Auto.
Generalize (interCorrectIncL M1 M'3 M'4 (app L'1 L'3) (app L'2 L'4));
 Rewrite EqM'5; Simpl; Auto.
Apply eqMemSym.
Apply eqExceptEq; Auto.
Apply eqExceptOnrTrans with M2 := M1; Auto.
Apply eqExceptOnIncl with L1 := L'5.
Generalize (interCorrectEq M1 M'3 M'4 (app L'1 L'3) (app L'2 L'4));
 Rewrite EqM'5; Auto.
Generalize (interCorrectIncR M1 M'3 M'4 (app L'1 L'3) (app L'2 L'4));
 Rewrite EqM'5; Simpl; Auto.
Apply eqExceptOnSym.
Apply eqExceptOnTrans with M2 := M'2; Auto.
Apply eqExceptOnSym.
Apply eqExceptOnTrans with M2 := M'1; Auto.
Apply eqExceptOnrTrans with M2 := M3; Auto.
Apply eqExceptOnRef; Auto.
Apply eqMemSym; Auto.
Apply eqExceptOnrTrans with M2 := M2; Auto.
Apply eqExceptOnRef; Auto.
Apply eqMemSym; Auto.
Qed.

Definition dilemmaF2aux:
  Mem -> Mem -> Mem -> Mem -> (list rZ) -> rZ -> rZ -> rZ -> rZ -> bool ->
  (((Mem * Mem) * Mem) * bool) * (list rZ) :=
   [M1, M2, M3, M4:Mem] [L:(list rZ)] [a, b, c, d:rZ] [b0:bool]
      Cases (add a b M1) of
          (pair (pair M'1 true) L'1) =>
            (((((reset M'1 M4 L'1), M2), M3), b0), L)
         | (pair (pair M'1 false) L'1) => Cases (add c d M'1) of
                                              (pair (pair M'2 true) L'2) =>
                                                (((((reset M'2 M4 (app L'1 L'2)),
                                                    M2), M3), b0), L)
                                             | (pair (pair M'2 false) L'2) =>
                                                 Cases (doTripletsF M'2 L'2) of
                                                     (exist
                                                        (pair
                                                           (pair M'3 b'3) L'3) _)
                                                       =>
                                                       (LetP
                                                          ? ?
                                                          (app
                                                             L'3 (app L'2 L'1))
                                                          [L'3L'2L'1:?] [_:?]
                                                             Cases b'3 of
                                                                 true =>
                                                                   (((((reset
                                                                          M'3 M4
                                                                          L'3L'2L'1),
                                                                       M2), M3),
                                                                     b0), L)
                                                                | false =>
                                                                    Cases b0 of
                                                                        true =>
                                                                          (((((reset
                                                                                 M2
                                                                                 M4
                                                                                 L),
                                                                              M'3),
                                                                             M3),
                                                                            false),
                                                                           L'3L'2L'1)
                                                                       | false
                                                                           =>
                                                                           Cases (inter
                                                                                    M3
                                                                                    M'3
                                                                                    M2
                                                                                    L'3L'2L'1
                                                                                    L) of
                                                                               (pair
                                                                                  M'5
                                                                                    L'5)
                                                                                 =>
                                                                                 (((((reset
                                                                                        M'3
                                                                                        M4
                                                                                        L'3L'2L'1),
                                                                                     M'5),
                                                                                    (reset
                                                                                       M2
                                                                                       M4
                                                                                       L)),
                                                                                   false),
                                                                                  L'5)
                                                                           end
                                                                    end
                                                             end)
                                                 end
                                          end
      end.

Definition dilemmaF2:
  Mem -> Mem -> Mem -> Mem -> rZ -> rZ ->
  (((Mem * Mem) * Mem) * bool) * (list rZ) :=
   [M1, M2, M3, M4:Mem] [a, b:rZ]
      Cases (dilemmaF2aux M1 M2 M3 M4 (nil ?) a rZTrue b rZTrue true) of
          (pair (pair (pair (pair M'1 M'2) M'3) b') L'1) =>
            Cases (dilemmaF2aux M'1 M'2 M'3 M4 L'1 a rZTrue b rZFalse b') of
                (pair (pair (pair (pair M''1 M''2) M''3) b'') L''1) =>
                  Cases (dilemmaF2aux
                           M''1 M''2 M''3 M4 L''1 a rZFalse b rZTrue b'') of
                      (pair (pair (pair (pair M'''1 M'''2) M'''3) b''') L'''1)
                        =>
                        (dilemmaF2aux
                           M'''1 M'''2 M'''3 M4 L'''1 a rZFalse b rZFalse b''')
                  end
            end
      end.

Theorem dilemma2FauxProp1:
  (M1, M2, M3, M4:Mem)
  (L:(list rZ))
  (b0:bool)
  (a, b, c, d:rZ) (eqMem M1 M4) -> (eqExceptOn M2 M4 L) -> (eqMem M3 M4) ->
  Cases (dilemmaF2aux M1 M2 M3 M4 L a b c d b0) of
      (pair (pair (pair (pair M'1 M'2) M'3) b') L'1) =>
        (eqMem M'1 M4) /\ ((eqExceptOn M'2 M4 L'1) /\ (eqMem M'3 M4))
  end.
Unfold dilemmaF2aux.
Intros M1 M2 M3 M4 L0 b0 a b c d eqM1M4 eqM2M4 eqM3M4.
Generalize (refl_equal ? (add a b M1)); Pattern -1 (add a b M1);
 Case (add a b M1).
Intros Mb; Case Mb; Clear Mb.
Intros M5 b5 L5; Case b5; Intros Eqab.
Repeat Split; Auto.
Apply eqExceptEq; Auto.
Apply eqExceptOnrTrans with M2 := M1; Auto.
Generalize (addCorrectEval M1 a b); Rewrite Eqab; Auto.
Generalize (refl_equal ? (add c d M5)); Pattern -1 (add c d M5);
 Case (add c d M5).
Intros Mb; Case Mb; Clear Mb.
Intros M6 b6 L6; Case b6; Intros Eqcd.
Repeat Split; Auto.
Apply eqExceptEq; Auto.
Apply eqExceptOnTrans with M2 := M5; Auto.
Generalize (addCorrectEval M5 c d); Rewrite Eqcd; Auto.
Apply eqExceptOnrTrans with M2 := M1; Auto.
Generalize (addCorrectEval M1 a b); Rewrite Eqab; Auto.
Case (doTripletsF M6 L6).
Intros x; Case x; Clear x.
Intros x; Case x; Clear x.
Unfold getMem LetP; Simpl.
Intros M7 b7 L7; Case b7.
Intros H'0; Elim H'0; Intros H'2 H'3; Elim H'3; Intros H'4 H'5; Repeat Split;
 Clear H'3 H'0; Auto.
Apply eqExceptEq; Auto.
Apply eqExceptOnSym.
Apply eqExceptOnTrans with M2 := M6; Auto.
Apply eqExceptOnTrans with M2 := M5; Auto.
Apply eqExceptOnrTrans with M2 := M1; Auto.
Apply eqExceptOnSym; Auto.
Generalize (addCorrectEval M1 a b); Rewrite Eqab; Auto.
Generalize (addCorrectEval M5 c d); Rewrite Eqcd; Auto.
Intros H'0; Elim H'0; Intros H'2 H'3; Elim H'3; Intros H'4 H'5; Case b0;
 Clear H'3 H'0.
Repeat Split; Auto.
Apply eqExceptEq; Auto.
Apply eqExceptOnSym.
Apply eqExceptOnTrans with M2 := M6; Auto.
Apply eqExceptOnTrans with M2 := M5; Auto.
Apply eqExceptOnrTrans with M2 := M1; Auto.
Apply eqExceptOnSym; Auto.
Generalize (addCorrectEval M1 a b); Rewrite Eqab; Auto.
Generalize (addCorrectEval M5 c d); Rewrite Eqcd; Auto.
Generalize (refl_equal ? (inter M3 M7 M2 (app L7 (app L6 L5)) L0));
 Pattern -1 (inter M3 M7 M2 (app L7 (app L6 L5)) L0);
 Case (inter M3 M7 M2 (app L7 (app L6 L5)) L0).
Intros M8 L8 eqInter; Repeat Split; Auto.
Apply eqExceptEq; Auto.
Apply eqExceptOnSym.
Apply eqExceptOnTrans with M2 := M6; Auto.
Apply eqExceptOnTrans with M2 := M5; Auto.
Apply eqExceptOnrTrans with M2 := M1; Auto.
Apply eqExceptOnSym; Auto.
Generalize (addCorrectEval M1 a b); Rewrite Eqab; Auto.
Generalize (addCorrectEval M5 c d); Rewrite Eqcd; Auto.
Apply eqExceptOnrTrans with M2 := M3; Auto.
Generalize (interCorrectEq M3 M7 M2 (app L7 (app L6 L5)) L0); Rewrite eqInter;
 Simpl; Auto.
Intros H'; Apply eqExceptOnSym; Auto.
Apply H'; Auto.
Apply eqExceptOnTrans with M2 := M6; Auto.
Apply eqExceptOnTrans with M2 := M5; Auto.
Apply eqExceptOnrTrans with M2 := M1; Auto.
Apply eqExceptOnSym; Auto.
Apply eqExceptOnrTrans with M2 := M4; Auto.
Apply eqExceptOnSym; Auto.
Generalize (addCorrectEval M1 a b); Rewrite Eqab; Auto.
Generalize (addCorrectEval M5 c d); Rewrite Eqcd; Auto.
Apply eqExceptOnrTrans with M2 := M4; Auto.
Apply eqExceptEq; Auto.
Qed.

Local memFb :=
   [Mb:Mem * bool]
      Cases Mb of
          (pair M true) => (cons ((rZPlus zero), (rZMinus zero)) (nil ?))
         | (pair M false) => (memF M)
      end.

Theorem dilemma2FauxProp2:
  (M1, M2, M3, M4:Mem)
  (L1:(list rZ))
  (b0:bool)
  (a, b, c, d:rZ) (eqMem M1 M4) -> (eqExceptOn M2 M4 L1) -> (eqMem M3 M4) ->
  Cases (dilemmaF2aux M1 M2 M3 M4 L1 a b c d b0) of
      (pair (pair (pair (pair M'1 M'2) M'3) b') L'1) => Cases b0 of
                                                            true =>
                                                              (doTripletsP
                                                                 (addEq
                                                                    (a, b)
                                                                    (addEq
                                                                       (c, d)
                                                                       (memF M1)))
                                                                 L
                                                                 (memFb
                                                                    (M'2, b')))
                                                           | false =>
                                                               b'=false /\
                                                               (Ex [S:State]
                                                               (doTripletsP
                                                                  (addEq
                                                                     (a, b)
                                                                     (addEq
                                                                        (c, d)
                                                                        (memF
                                                                           M1)))
                                                                  L S) /\
                                                               (eqState
                                                                  (memF M'2)
                                                                  (interState
                                                                     (memF M2) S)))
                                                        end
  end.
Unfold dilemmaF2aux.
Intros M1 M2 M3 M4 L0 b0 a b c d eqM1M4 eqM2M4 eqM3M4.
Generalize (refl_equal ? (add a b M1)); Pattern -1 (add a b M1);
 Case (add a b M1).
Intros Mb; Case Mb; Clear Mb.
Intros M5 b5 L5; Case b5; Intros Eqab.
Case b0; Auto.
Apply doTripletsComp
     with S1 := (cons ((rZPlus zero), (rZMinus zero)) (nil ?))
          S2 := (cons ((rZPlus zero), (rZMinus zero)) (nil ?)); Auto.
Red; Split; Auto.
Apply inclStateTrans with S2 := (addEq (a, b) (memF M1)); Auto.
Red; Auto.
Intros i j H'.
Apply eqStateContr with p := a q := b; Auto.
Unfold addEq; Apply eqStateRzTail; Auto.
Apply evalInvComp; Auto.
Apply addCorrectContrEval; Auto.
Rewrite Eqab; Auto.
Repeat Split; Auto.
Exists (cons ((rZPlus zero), (rZMinus zero)) (nil ?)); Split; Auto.
Apply doTripletsComp
     with S1 := (cons ((rZPlus zero), (rZMinus zero)) (nil ?))
          S2 := (cons ((rZPlus zero), (rZMinus zero)) (nil ?)); Auto.
Red; Repeat Split; Auto.
Apply inclStateTrans with S2 := (addEq (a, b) (memF M1)); Auto.
Red; Auto.
Intros i j H'.
Apply eqStateContr with p := a q := b; Auto.
Unfold addEq; Apply eqStateRzTail; Auto.
Apply evalInvComp; Auto.
Apply addCorrectContrEval; Auto.
Rewrite Eqab; Auto.
Generalize (refl_equal ? (add c d M5)); Pattern -1 (add c d M5);
 Case (add c d M5).
Intros Mb; Case Mb; Clear Mb.
Intros M6 b6 L6; Case b6; Intros Eqcd.
Case b0; Auto.
Apply doTripletsComp
     with S1 := (cons ((rZPlus zero), (rZMinus zero)) (nil ?))
          S2 := (cons ((rZPlus zero), (rZMinus zero)) (nil ?)); Auto.
Red; Repeat Split; Auto.
Apply inclStateEqStateComp
     with S1 := (cons ((rZPlus zero), (rZMinus zero)) (nil ?))
          S3 := (addEq (c, d) (addEq (a, b) (memF M1))); Auto.
Red; Auto.
Intros i j H'.
Apply eqStateContr with p := c q := d; Auto.
Unfold addEq; Apply eqStateRzTail; Auto.
Apply eqStateEq with S1 := (memF M5); Auto.
Generalize (addCorrectExEq M1 a b).
Rewrite Eqab; Auto.
Apply evalInvComp.
Apply addCorrectContrEval; Auto.
Rewrite Eqcd; Auto.
Repeat Split; Auto.
Exists (cons ((rZPlus zero), (rZMinus zero)) (nil ?)); Split; Auto.
Apply doTripletsComp
     with S1 := (cons ((rZPlus zero), (rZMinus zero)) (nil ?))
          S2 := (cons ((rZPlus zero), (rZMinus zero)) (nil ?)); Auto.
Red; Repeat Split; Auto.
Apply inclStateTrans with S2 := (addEq (c, d) (addEq (a, b) (memF M1))); Auto.
Red; Auto.
Intros i j H'.
Apply eqStateContr with p := c q := d; Auto.
Unfold addEq; Apply eqStateRzTail; Auto.
Apply eqStateEq with S1 := (memF M5); Auto.
Generalize (addCorrectExEq M1 a b).
Rewrite Eqab; Auto.
Apply evalInvComp.
Apply addCorrectContrEval; Auto.
Rewrite Eqcd; Auto.
Case (doTripletsF M6 L6).
Intros x; Case x; Clear x.
Intros x; Case x; Clear x.
Unfold getMem LetP; Simpl.
Intros M7 b7 L7; Case b7.
Intros H'0; Elim H'0; Intros H'2 H'3; Elim H'3; Intros H'4 H'5; Repeat Split;
 Clear H'3 H'0; Auto.
Case b0; Auto.
Apply doTripletsComp
     with S1 := (memF M6) S2 := (cons ((rZPlus zero), (rZMinus zero)) (nil ?));
 Auto.
Apply eqStateTrans with S2 := (addEq (c, d) (addEq (a, b) (memF M1))); Auto.
Apply eqStateTrans with S2 := (addEq (c, d) (memF M5)); Auto.
Apply eqStateAddEq; Auto.
Generalize (addCorrectExEq M1 a b).
Rewrite Eqab; Auto.
Generalize (addCorrectExEq M5 c d).
Rewrite Eqcd; Auto.
Repeat Split; Auto.
Exists (cons ((rZPlus zero), (rZMinus zero)) (nil ?)); Split; Auto.
Apply doTripletsComp
     with S1 := (memF M6) S2 := (cons ((rZPlus zero), (rZMinus zero)) (nil ?));
 Auto.
Apply eqStateTrans with S2 := (addEq (c, d) (addEq (a, b) (memF M1))); Auto.
Apply eqStateTrans with S2 := (addEq (c, d) (memF M5)); Auto.
Apply eqStateAddEq; Auto.
Generalize (addCorrectExEq M1 a b).
Rewrite Eqab; Auto.
Generalize (addCorrectExEq M5 c d).
Rewrite Eqcd; Auto.
Intros H'; Elim H'; Intros H'0 H'1; Elim H'1; Intros H'2 H'3; Clear H'1 H'.
Case b0; Auto.
Apply doTripletsComp with S1 := (memF M6) S2 := (memF M7); Auto.
Apply eqStateTrans with S2 := (addEq (c, d) (addEq (a, b) (memF M1))); Auto.
Apply eqStateTrans with S2 := (addEq (c, d) (memF M5)); Auto.
Apply eqStateAddEq; Auto.
Generalize (addCorrectExEq M1 a b).
Rewrite Eqab; Auto.
Generalize (addCorrectExEq M5 c d).
Rewrite Eqcd; Auto.
Generalize (refl_equal ? (inter M3 M7 M2 (app L7 (app L6 L5)) L0));
 Pattern -1 (inter M3 M7 M2 (app L7 (app L6 L5)) L0);
 Case (inter M3 M7 M2 (app L7 (app L6 L5)) L0).
Intros M8 L8 eqInter; Repeat Split; Auto.
Exists (memF M7); Split; Auto.
Apply doTripletsComp with S1 := (memF M6) S2 := (memF M7); Auto.
Apply eqStateTrans with S2 := (addEq (c, d) (addEq (a, b) (memF M1))); Auto.
Apply eqStateTrans with S2 := (addEq (c, d) (memF M5)); Auto.
Apply eqStateAddEq; Auto.
Generalize (addCorrectExEq M1 a b).
Rewrite Eqab; Auto.
Generalize (addCorrectExEq M5 c d).
Rewrite Eqcd; Auto.
Apply eqStateTrans with S2 := (interState (memF M7) (memF M2)); Auto.
Generalize (interCorrect M3 M7 M2 (app L7 (app L6 L5)) L0);
 (Rewrite eqInter; Simpl; Auto).
Intros H'; Apply H'; Auto.
Apply eqExceptOnTrans with M2 := M6; Auto.
Apply eqExceptOnTrans with M2 := M5; Auto.
Apply eqExceptOnrTrans with M2 := M1; Auto.
Apply eqExceptOnrTrans with M2 := M4; Auto.
Apply eqExceptOnSym; Auto.
Generalize (addCorrectEval M1 a b); Rewrite Eqab; Auto.
Generalize (addCorrectEval M5 c d); Rewrite Eqcd; Auto.
Apply eqExceptOnrTrans with M2 := M4; Auto.
Qed.

Theorem dilemma2FProp1:
  (M1, M2, M3, M4:Mem)
  (a, b:rZ) (eqMem M1 M4) -> (eqMem M2 M4) -> (eqMem M3 M4) ->
  Cases (dilemmaF2 M1 M2 M3 M4 a b) of
      (pair (pair (pair (pair M'1 M'2) M'3) b') L'1) =>
        (eqMem M'1 M4) /\ ((eqExceptOn M'2 M4 L'1) /\ (eqMem M'3 M4))
  end.
Intros M1 M2 M3 M4 a b EqM1M4 EqM2M4 EqM3M4; Unfold dilemmaF2.
Generalize (dilemma2FauxProp1 M1 M2 M3 M4 (nil rZ) true a rZTrue b rZTrue).
Case (dilemmaF2aux M1 M2 M3 M4 (nil rZ) a rZTrue b rZTrue true).
Intros p; Case p; Clear p.
Intros p; Case p; Clear p.
Intros p; Case p; Clear p.
Simpl; Auto.
Intros M'1 M'2 M'3 b' L'1.
Intros H'; Elim H';
 [Intros H'3 H'4; Elim H'4; Intros H'5 H'6; Clear H'4 H' | Clear H' | Clear H' |
   Clear H']; Auto.
Generalize (dilemma2FauxProp1 M'1 M'2 M'3 M4 L'1 b' a rZTrue b rZFalse).
Case (dilemmaF2aux M'1 M'2 M'3 M4 L'1 a rZTrue b rZFalse b').
Intros p; Case p; Clear p.
Intros p; Case p; Clear p.
Intros p; Case p; Clear p.
Simpl; Auto.
Intros M''1 M''2 M''3 b'' L''1.
Intros H'; Elim H';
 [Intros H'4 H'7; Elim H'7; Intros H'8 H'9; Clear H'7 H' | Clear H' | Clear H' |
   Clear H']; Auto.
Generalize (dilemma2FauxProp1 M''1 M''2 M''3 M4 L''1 b'' a rZFalse b rZTrue).
Case (dilemmaF2aux M''1 M''2 M''3 M4 L''1 a rZFalse b rZTrue b'').
Intros p; Case p; Clear p.
Intros p; Case p; Clear p.
Intros p; Case p; Clear p.
Simpl; Auto.
Intros M'''1 M'''2 M'''3 b''' L'''1.
Intros H'; Elim H';
 [Intros H'7 H'10; Elim H'10; Intros H'11 H'12; Clear H'10 H' | Clear H' |
   Clear H' | Clear H']; Auto.
Generalize (dilemma2FauxProp1
              M'''1 M'''2 M'''3 M4 L'''1 b''' a rZFalse b rZFalse).
Case (dilemmaF2aux M'''1 M'''2 M'''3 M4 L'''1 a rZFalse b rZFalse b''').
Intros p; Case p; Clear p.
Intros p; Case p; Clear p.
Intros p; Case p; Clear p.
Simpl; Auto.
Qed.

Theorem dilemma2FProp2Aux:
  (S1, S2, S3, S4, S5, S6, S7, S8, S9:State)
  (a, b:rZ)
  (doTripletsP (addEq (a, rZTrue) (addEq (b, rZTrue) S1)) L S5) ->
  (doTripletsP (addEq (a, rZTrue) (addEq (b, rZFalse) S2)) L S6) ->
  (doTripletsP (addEq (a, rZFalse) (addEq (b, rZTrue) S3)) L S7) ->
  (doTripletsP (addEq (a, rZFalse) (addEq (b, rZFalse) S4)) L S8) ->
  (eqState S1 S2) ->
  (eqState S1 S3) ->
  (eqState S1 S4) ->
  (eqState (interState (interState S5 S7) (interState S6 S8)) S9) ->
  (dilemmaP S1 L S9).
Intros S1 S2 S3 S4 S5 S6 S7 S8 S9 a b H' H'0 H'1 H'2 H'3 H'4 H'5 H'6.
Apply dilemmaPSplit
     with a := b b := rZTrue S2 := (interState S5 S7) S3 := (interState S6 S8);
 Auto.
Apply dilemmaPSplit with a := a b := rZTrue S2 := S5 S3 := S7; Auto.
Apply dilemmaComp with 1 := (dilemmaPRef ? ? ? H'1); Auto.
Apply dilemmaPSplit with a := a b := rZTrue S2 := S6 S3 := S8; Auto.
Apply dilemmaComp with 1 := (dilemmaPRef ? ? ? H'0); Auto.
Apply dilemmaComp with 1 := (dilemmaPRef ? ? ? H'2); Auto.
Qed.

Theorem dilemma2FProp2:
  (M1, M2, M3, M4:Mem)
  (a, b:rZ) (eqMem M1 M4) -> (eqMem M2 M4) -> (eqMem M3 M4) ->
  Cases (dilemmaF2 M1 M2 M3 M4 a b) of
      (pair (pair (pair (pair M'1 M'2) M'3) b') L'1) =>
        (dilemmaP (memF M1) L (memFb (M'2, b')))
  end.
Intros M1 M2 M3 M4 a b EqM1M4 EqM2M4 EqM3M4; Unfold dilemmaF2.
Generalize (dilemma2FauxProp2 M1 M2 M3 M4 (nil rZ) true a rZTrue b rZTrue).
Generalize (dilemma2FauxProp1 M1 M2 M3 M4 (nil rZ) true a rZTrue b rZTrue).
Case (dilemmaF2aux M1 M2 M3 M4 (nil rZ) a rZTrue b rZTrue true).
Intros p; Case p; Clear p.
Intros p; Case p; Clear p.
Intros p; Case p; Clear p.
Simpl; Auto.
Intros M'1 M'2 M'3 b' L'1.
Intros H'; Elim H';
 [Intros H'3 H'4; Elim H'4; Intros H'5 H'6; Clear H'4 H' | Clear H' | Clear H' |
   Clear H']; Auto.
Intros H'; LApply H';
 [Intros H'0; LApply H'0;
   [Intros H'1; LApply H'1; [Clear H'1 H'0 H' | Clear H'1 H'0 H'] | Clear H'0 H'] |
   Clear H']; Auto.
Generalize (dilemma2FauxProp2 M'1 M'2 M'3 M4 L'1 b' a rZTrue b rZFalse).
Generalize (dilemma2FauxProp1 M'1 M'2 M'3 M4 L'1 b' a rZTrue b rZFalse).
Case (dilemmaF2aux M'1 M'2 M'3 M4 L'1 a rZTrue b rZFalse b').
Intros p; Case p; Clear p.
Intros p; Case p; Clear p.
Intros p; Case p; Clear p.
Simpl; Auto.
Intros M''1 M''2 M''3 b'' L''1.
Intros H'; Elim H';
 [Intros H'4 H'7; Elim H'7; Intros H'8 H'9; Clear H'7 H' | Clear H' | Clear H' |
   Clear H']; Auto.
Intros H'; LApply H';
 [Intros H'0; LApply H'0;
   [Intros H'1; LApply H'1; [Clear H'1 H'0 H' | Clear H'1 H'0 H'] | Clear H'0 H'] |
   Clear H']; Auto.
Generalize (dilemma2FauxProp2 M''1 M''2 M''3 M4 L''1 b'' a rZFalse b rZTrue).
Generalize (dilemma2FauxProp1 M''1 M''2 M''3 M4 L''1 b'' a rZFalse b rZTrue).
Case (dilemmaF2aux M''1 M''2 M''3 M4 L''1 a rZFalse b rZTrue b'').
Intros p; Case p; Clear p.
Intros p; Case p; Clear p.
Intros p; Case p; Clear p.
Simpl; Auto.
Intros M'''1 M'''2 M'''3 b''' L'''1.
Intros H'; Elim H';
 [Intros H'7 H'10; Elim H'10; Intros H'11 H'12; Clear H'10 H' | Clear H' |
   Clear H' | Clear H']; Auto.
Intros H'; LApply H';
 [Intros H'0; LApply H'0;
   [Intros H'1; LApply H'1; [Try Exact H'2; Clear H'1 H'0 H' | Clear H'1 H'0 H'] |
     Clear H'0 H'] | Clear H']; Auto.
Generalize (dilemma2FauxProp2
              M'''1 M'''2 M'''3 M4 L'''1 b''' a rZFalse b rZFalse).
Generalize (dilemma2FauxProp1
              M'''1 M'''2 M'''3 M4 L'''1 b''' a rZFalse b rZFalse).
Case (dilemmaF2aux M'''1 M'''2 M'''3 M4 L'''1 a rZFalse b rZFalse b''').
Intros p; Case p; Clear p.
Intros p; Case p; Clear p.
Intros p; Case p; Clear p.
Simpl; Auto.
Intros M''''1 M''''2 M''''3 b'''' L''''1 H'; Elim H';
 [Intros H'10 H'13; Elim H'13; Intros H'14 H'15; Clear H'13 H' | Clear H' |
   Clear H' | Clear H']; Auto.
Intros H'; LApply H';
 [Intros H'0; LApply H'0;
   [Intros H'1; LApply H'1; [Clear H'1 H'0 H' | Clear H'1 H'0 H'] | Clear H'0 H'] |
   Clear H']; Auto.
Cut (eqState (memF M1) (memF M'1));
 [Intros M10 |
   Apply eqqMemImpEqState; Apply eqMemTrans with y := M4; Auto; Apply eqMemSym];
 Auto.
Cut (eqState (memF M1) (memF M''1));
 [Intros M11 |
   Apply eqqMemImpEqState; Apply eqMemTrans with y := M4; Auto; Apply eqMemSym];
 Auto.
Cut (eqState (memF M1) (memF M'''1));
 [Intros M12 |
   Apply eqqMemImpEqState; Apply eqMemTrans with y := M4; Auto; Apply eqMemSym];
 Auto.
Cut (eqState (memF M1) (memF M''''1));
 [Intros M13 |
   Apply eqqMemImpEqState; Apply eqMemTrans with y := M4; Auto; Apply eqMemSym];
 Auto.
Case b'''; Intros FB;
 [Idtac |
   Elim FB; Intros H' H'0; Elim H'0; Clear H'0 FB; Intros SF E; Elim E;
    Intros FB EqFB; Clear E]; Auto;
 (Case b''; Intros SB;
   [Idtac |
     Elim SB; Clear SB; Intros Disc H'0; Elim H'0; Intros SS E; Elim E;
      Clear E H'0; Intros SB EqSB; Try Discriminate Disc; Clear Disc]); Auto;
 (Case b'; Intros TB;
   [Idtac |
     Elim TB; Clear TB; Intros Disc H'0; Elim H'0; Intros ST E; Elim E;
      Clear E H'0; Intros TB EqTB; Try Discriminate Disc; Clear Disc]);
 Intros QB; Try Rewrite H';
 Apply dilemma2FProp2Aux with 1 := QB 2 := TB 3 := SB 4 := FB; Auto.
Apply eqStateTrans
     with S2 :=
          (interState
             (cons ((rZPlus zero), (rZMinus zero)) (nil rZ * rZ))
             (Case b'''' of
               (cons ((rZPlus zero), (rZMinus zero)) (nil rZ * rZ))
               (memF M''''2) end)); Auto.
Apply eqStateTrans with S2 := (interState (memF M'''2) SF); Auto.
Apply eqStateTrans with S2 := (interState SS (interState (memF M''2) SF)); Auto.
Apply eqStateTrans with S2 := (interState (memF M'''2) SF); Auto.
Apply eqStateTrans with S2 := (interState (interState (memF M''2) SS) SF); Auto.
Apply eqStateTrans with S2 := (interState (interState SS (memF M''2)) SF); Auto.
Apply eqStateTrans with S2 := (interState (memF M'''2) SF); Auto.
Apply eqStateTrans with S2 := (interState (interState (memF M''2) SS) SF); Auto.
Apply eqStateTrans
     with S2 := (interState (interState (interState (memF M'2) ST) SS) SF); Auto.
Apply eqStateTrans
     with S2 := (interState (interState (interState (memF M'2) SS) ST) SF); Auto.
Apply interStateEq; Auto.
Apply eqStateTrans with S2 := (interState (memF M'2) (interState SS ST)); Auto.
Apply eqStateTrans with S2 := (interState (memF M'2) (interState ST SS)); Auto.
Qed.

Theorem dilemma2FauxProp3:
  (M1, M2, M3, M4:Mem)
  (L:(list rZ))
  (b0:bool)
  (a, b, c, d:rZ) (ltWithList M2 M4 L) -> (eqMem M1 M4) -> (eqMem M1 M3) ->
  Cases (dilemmaF2aux M1 M2 M3 M4 L a b c d b0) of
      (pair (pair (pair (pair M'1 M'2) M'3) b') L'1) => (ltWithList M'2 M4 L'1)
  end.
Unfold dilemmaF2aux.
Intros M1 M2 M3 M4 L0 b0 a b c d ltM2M4 eqM1M4 eqM1M3.
Generalize (refl_equal ? (add a b M1)); Pattern -1 (add a b M1);
 Case (add a b M1).
Intros Mb; Case Mb; Clear Mb.
Intros M5 b5 L5; Case b5; Intros Eqab.
Repeat Split; Auto.
Generalize (refl_equal ? (add c d M5)); Pattern -1 (add c d M5);
 Case (add c d M5).
Intros Mb; Case Mb; Clear Mb.
Intros M6 b6 L6; Case b6; Intros Eqcd.
Repeat Split; Auto.
Case (doTripletsF M6 L6).
Intros x; Case x; Clear x.
Intros x; Case x; Clear x.
Unfold getMem LetP; Simpl.
Intros M7 b7 L7; Case b7; Auto.
Intros H'0; Elim H'0; Intros H'2 H'3; Elim H'3; Intros H'4 H'5; Case b0;
 Clear H'3 H'0.
Rewrite appSym; Auto.
Apply ltWithListTrans with M2 := M6; Auto.
Rewrite appSym; Auto.
Apply ltWithListTrans with M2 := M5; Auto.
Generalize (addCorrectlt M5 c d); Auto; Rewrite Eqcd; Auto.
Apply ltWithEq with M1 := M5 M3 := M1; Auto.
Generalize (addCorrectlt M1 a b); Auto; Rewrite Eqab; Auto.
Generalize (refl_equal ? (inter M3 M7 M2 (app L7 (app L6 L5)) L0));
 Pattern -1 (inter M3 M7 M2 (app L7 (app L6 L5)) L0);
 Case (inter M3 M7 M2 (app L7 (app L6 L5)) L0).
Intros M8 L8 eqInter; Auto.
Apply ltWithEq with M1 := M8 M3 := M3; Auto.
Apply eqMemTrans with y := M1; Auto.
Apply eqMemSym; Auto.
Generalize (interlt M3 M7 M2 (app L7 (app L6 L5)) L0); Auto; Rewrite eqInter;
 Simpl; Auto.
Qed.

Theorem dilemma2FProp3:
  (M1, M2, M3, M4:Mem)
  (a, b:rZ) (eqMem M1 M4) -> (eqMem M2 M4) -> (eqMem M3 M4) ->
  Cases (dilemmaF2 M1 M2 M3 M4 a b) of
      (pair (pair (pair (pair M'1 M'2) M'3) b') L'1) => (ltWithList M'2 M4 L'1)
  end.
Intros M1 M2 M3 M4 a b EqM1M4 EqM2M4 EqM3M4; Unfold dilemmaF2.
Generalize (dilemma2FauxProp3 M1 M2 M3 M4 (nil rZ) true a rZTrue b rZTrue).
Generalize (dilemma2FauxProp1 M1 M2 M3 M4 (nil rZ) true a rZTrue b rZTrue).
Case (dilemmaF2aux M1 M2 M3 M4 (nil rZ) a rZTrue b rZTrue true).
Intros p; Case p; Clear p.
Intros p; Case p; Clear p.
Intros p; Case p; Clear p.
Simpl; Auto.
Intros M'1 M'2 M'3 b' L'1.
Intros H'; Elim H';
 [Intros H'3 H'4; Elim H'4; Intros H'5 H'6; Clear H'4 | Idtac | Idtac | Idtac];
 Clear H'; Auto; Intros H'.
Generalize (dilemma2FauxProp3 M'1 M'2 M'3 M4 L'1 b' a rZTrue b rZFalse).
Generalize (dilemma2FauxProp1 M'1 M'2 M'3 M4 L'1 b' a rZTrue b rZFalse).
Case (dilemmaF2aux M'1 M'2 M'3 M4 L'1 a rZTrue b rZFalse b').
Intros p; Case p; Clear p.
Intros p; Case p; Clear p.
Intros p; Case p; Clear p.
Simpl; Auto.
Intros M''1 M''2 M''3 b'' L''1.
Intros H'0; Elim H'0;
 [Intros H'7 H'8; Elim H'8; Intros H'9 H'10; Clear H'8 | Idtac | Idtac | Idtac];
 Clear H'0; Auto; Intros H'0.
Generalize (dilemma2FauxProp3 M''1 M''2 M''3 M4 L''1 b'' a rZFalse b rZTrue).
Generalize (dilemma2FauxProp1 M''1 M''2 M''3 M4 L''1 b'' a rZFalse b rZTrue).
Case (dilemmaF2aux M''1 M''2 M''3 M4 L''1 a rZFalse b rZTrue b'').
Intros p; Case p; Clear p.
Intros p; Case p; Clear p.
Intros p; Case p; Clear p.
Simpl; Auto.
Intros M'''1 M'''2 M'''3 b''' L'''1.
Intros H'1; Elim H'1;
 [Intros H'11 H'12; Elim H'12; Intros H'13 H'14; Clear H'12 | Idtac | Idtac |
   Idtac]; Clear H'1; Auto; Intros H'1.
Generalize (dilemma2FauxProp3
              M'''1 M'''2 M'''3 M4 L'''1 b''' a rZFalse b rZFalse).
Generalize (dilemma2FauxProp1
              M'''1 M'''2 M'''3 M4 L'''1 b''' a rZFalse b rZFalse).
Case (dilemmaF2aux M'''1 M'''2 M'''3 M4 L'''1 a rZFalse b rZFalse b''').
Intros p; Case p; Clear p.
Intros p; Case p; Clear p.
Intros p; Case p; Clear p.
Intros m m0 m1 H'2 l H'4; Elim H'4;
 [Intros H'16 H'17; Elim H'17; Intros H'18 H'19; Clear H'17 | Idtac | Idtac |
   Idtac]; Clear H'4; Auto; Intros H'4.
Apply H'4; Auto.
Apply H'1; Auto.
Apply H'0; Auto.
Apply H'; Auto.
Apply eqMemTrans with y := M4; Auto.
Apply eqMemSym; Auto.
Apply eqMemTrans with y := M4; Auto.
Apply eqMemSym; Auto.
Apply eqMemTrans with y := M4; Auto.
Apply eqMemSym; Auto.
Apply eqMemTrans with y := M4; Auto.
Apply eqMemSym; Auto.
Qed.
End algo.
