Require Export stalmarck.
Require Export LetP.
Require Export Option.
Require Export state.
Require Export restrictState.
Require Export memoryImplement.
Require Export addEqImplement.
Require Export addEqImplement2.
Require Export OrderedListEq.
Require Export triplet.
Require Export state.
Require Export ltState.
Require Export array.
Require Export ltState.
Section algo.
Variable LL:(list triplet).
Variable maxN:rNat.
Hypothesis HmaxN:maxN=(rnext (maxVarTriplets LL)).

Definition memFx: ((Mem maxN) * bool) * (dList maxN) ->State :=
   [L:?]
      Cases L of
          (pair (pair M true) L) =>
            (cons ((rZPlus zero), (rZMinus zero)) (nil ?))
         | (pair (pair M false) L) => (memF maxN M)
      end.

Theorem addCorrectExEq:
  (M:(Mem maxN)) (a, b:rZ) (Ha:(rVlt a maxN)) (Hb:(rVlt b maxN))
  (eqState (addEq (a, b) (memF maxN M)) (memFx (addEqMem maxN M a b Ha Hb))).
Intros M a b Ha Hb; Generalize (addEqMemProp26 maxN M a b Ha Hb);
 Generalize (addEqMemProp28 maxN M a b Ha Hb).
Case (addEqMem maxN M a b Ha Hb); Simpl; Auto.
Intros p; Case p; Auto.
Intros m b0; Case b0; Auto.
Intros H' H'0 H'1; Red; Split; Auto.
Apply inclStateIn; Simpl; Auto.
Intros a0 b1 H'2; Elim H'2; Clear H'2; Intros H'3; Inversion H'3; Auto.
Apply eqStateContr with p := a q := b; Auto.
Apply eqStateaddEq2; Auto.
Apply evalInvComp with Maxa := Ha Maxb := Hb; Auto.
Qed.

Theorem addEqMemIncl:
  (M:(Mem maxN)) (a, b:rZ) (Ha:(rVlt a maxN)) (Hb:(rVlt b maxN))
  (inclState
     (memF maxN M) (memF maxN (getMem maxN (addEqMem maxN M a b Ha Hb)))).
Intros M a b Ha Hb; Generalize (addEqMemProp26 maxN M a b Ha Hb);
 Generalize (addEqMemProp2 maxN M a b Ha Hb).
Case (addEqMem maxN M a b Ha Hb); Auto.
Intros p; Case p.
Intros m b0; Case b0; Simpl; Auto.
Intros d H'; Elim H'; Intros H'0 H'1; Rewrite <- H'0; Clear H'; Auto.
Intros H' H'0 H'1.
Apply inclStateEqStateComp
     with S1 := (memF maxN M) S3 := (addEq (a, b) (memF maxN M)); Auto.
Qed.
Hints Resolve addEqMemIncl.

Theorem addEqMemNEq:
  (M:(Mem maxN)) (a, b:rZ) (Ha:(rVlt a maxN)) (Hb:(rVlt b maxN))
  (nEq
     maxN M (getMem maxN (addEqMem maxN M a b Ha Hb))
     (getList maxN (addEqMem maxN M a b Ha Hb))).
Intros M a b Ha Hb; Unfold nEq; Generalize (addEqMemProp27 maxN M a b Ha Hb).
Case (addEqMem maxN M a b Ha Hb); Auto.
Intros p; Case p; Auto.
Intros m b0; Simpl; Auto.
Intros d H' H'0; Red; Intros H'1.
LApply H'; Clear H'; [Intros H'2; Case H'2 | Idtac].
Apply eqMemEqState; Auto.
Red; Intros H'; Absurd (notDnil maxN d); Auto; Rewrite H'; Auto.
Qed.
Hints Resolve addEqMemNEq.

Definition add2EqMem :=
   [M:(Mem maxN)]
   [a, b, c, d:rZ]
   [Ha:(rVlt a maxN)] [Hb:(rVlt b maxN)] [Hc:(rVlt c maxN)] [Hd:(rVlt d maxN)]
      Cases (addEqMem maxN M c d Hc Hd) of
          (pair (pair M' b') L') => Cases b' of
                                        true => ((M', b'), L')
                                       | false =>
                                           Cases (addEqMem maxN M' a b Ha Hb) of
                                               (pair (pair M'' b'') L'') =>
                                                 ((M'', b''),
                                                  (appd maxN L' L''))
                                           end
                                    end
      end.

Theorem addEq2SumEq:
  (S:State) (a, b, c, d:rZ)
  (eqState (addEq (a, b) (addEq (c, d) S)) (addEq (c, d) (addEq (a, b) S))).
Intros S a b c d.
Apply eqStateTrans
     with S2 :=
          (unionState
             (unionState (cons (c, d) (nil ?)) (cons (a, b) (nil ?))) S); Auto.
Apply eqStateTrans
     with S2 :=
          (unionState
             (unionState (cons (a, b) (nil ?)) (cons (c, d) (nil ?))) S); Auto.
Qed.
Hints Resolve addEq2SumEq.

Theorem addEqMemExcept:
  (M:(Mem maxN)) (p, q:rZ) (Hp:(rVlt p maxN)) (Hq:(rVlt q maxN))
  (eqExceptOn
     maxN M (getMem maxN (addEqMem maxN M p q Hp Hq))
     (getList maxN (addEqMem maxN M p q Hp Hq))).
Intros M p q Hp Hq; Generalize (addEqMemProp18 maxN M p q Hp Hq).
Case (addEqMem maxN M p q Hp Hq); Auto.
Intros p0 d; Case p0; Simpl; Auto.
Intros m H' H'0; Apply eqExceptOnSym; Auto.
Qed.
Hints Resolve addEqMemExcept.

Theorem add2EqMemExcept:
  (M:(Mem maxN))
  (p, q, r, s:rZ)
  (Hp:(rVlt p maxN)) (Hq:(rVlt q maxN)) (Hr:(rVlt r maxN)) (Hs:(rVlt s maxN))
  (eqExceptOn
     maxN M (getMem maxN (add2EqMem M p q r s Hp Hq Hr Hs))
     (getList maxN (add2EqMem M p q r s Hp Hq Hr Hs))).
Intros M p q r s Hp Hq Hr Hs; Unfold add2EqMem.
Generalize (addEqMemExcept M r s Hr Hs).
Case (addEqMem maxN M r s Hr Hs); Auto.
Intros p0; Case p0; Auto.
Intros m b; Case b; Auto.
Intros d H'.
Generalize (addEqMemExcept m p q Hp Hq).
Generalize (addEqMem maxN m p q Hp Hq); Auto.
Intros p1; Case p1; Auto.
Intros p2; Case p2; Simpl; Auto.
Intros m0 H'0 d0 H'1.
Apply eqExceptOnAppd with M2 := m; Auto.
Qed.
Hints Resolve add2EqMemExcept.

Theorem add2EqMemIncl:
  (M:(Mem maxN))
  (p, q, r, s:rZ)
  (Hp:(rVlt p maxN)) (Hq:(rVlt q maxN)) (Hr:(rVlt r maxN)) (Hs:(rVlt s maxN))
  (inclState
     (memF maxN M) (memF maxN (getMem maxN (add2EqMem M p q r s Hp Hq Hr Hs)))).
Intros M p q r s Hp Hq Hr Hs; Unfold add2EqMem.
Generalize (addEqMemIncl M r s Hr Hs).
Case (addEqMem maxN M r s Hr Hs); Auto.
Intros p0; Case p0; Auto.
Intros m b; Case b; Simpl; Auto.
Intros d.
Generalize (addEqMemIncl m p q Hp Hq).
Case (addEqMem maxN m p q Hp Hq); Auto.
Intros p1; Case p1; Simpl; Auto.
Intros m0 H' H'0 H'1 H'2.
Apply inclStateTrans with S2 := (memF maxN m); Auto.
Qed.
Hints Resolve add2EqMemIncl.

Theorem add2EqMemNEq:
  (M:(Mem maxN))
  (p, q, r, s:rZ)
  (Hp:(rVlt p maxN)) (Hq:(rVlt q maxN)) (Hr:(rVlt r maxN)) (Hs:(rVlt s maxN))
  (nEq
     maxN M (getMem maxN (add2EqMem M p q r s Hp Hq Hr Hs))
     (getList maxN (add2EqMem M p q r s Hp Hq Hr Hs))).
Intros M p q r s Hp Hq Hr Hs; Unfold add2EqMem.
Generalize (addEqMemNEq M r s Hr Hs); Generalize (addEqMemIncl M r s Hr Hs).
Case (addEqMem maxN M r s Hr Hs); Auto.
Intros p0; Case p0; Auto.
Intros m b; Case b; Simpl; Auto.
Generalize (addEqMemNEq m p q Hp Hq); Generalize (addEqMemIncl m p q Hp Hq).
Case (addEqMem maxN m p q Hp Hq); Simpl; Auto.
Intros p1; Case p1; Auto.
Intros m0 b0; Simpl; Auto.
Intros d H' H'0 d0 H'1 H'2.
Apply nEqAppd with M2 := m; Auto.
Qed.
Hints Resolve add2EqMemNEq.

Theorem doTripletFAux1:
  (M:(Mem maxN))
  (t:triplet)
  (p, p', q, q':rZ)
  (Hp:(rVlt p maxN))
  (Hp':(rVlt p' maxN))
  (Hq:(rVlt q maxN))
  (Hq':(rVlt q' maxN))
  p'=(evalMZ maxN M p Hp) ->
  q'=(evalMZ maxN M q Hq) ->
  (doTripletP (memF maxN M) t (addEq (p, q) (memF maxN M))) ->
  (ex
     State
     [S:State]
     (doTripletP (memF maxN M) t S) /\
     (eqState S (memFx (addEqMem maxN M p' q' Hp' Hq')))).
Intros M t p p' q q' Hp Hp' Hq Hq' H' H'0 H'1;
 Exists (addEq (p, q) (memF maxN M)); Split; Auto.
Apply eqStateTrans
     with S2 :=
          (addEq ((evalMZ maxN M p Hp), (evalMZ maxN M q Hq)) (memF maxN M));
 Auto.
Apply addEqComp; Auto.
Apply evalEq; Auto.
Apply evalEq; Auto.
Rewrite <- H'; Auto.
Rewrite <- H'0; Auto.
Apply addCorrectExEq; Auto.
Qed.
Hints Resolve doTripletFAux1.

Theorem doTripletFAux2:
  (M:(Mem maxN))
  (t:triplet)
  (p, p', q, q', r, r', s, s':rZ)
  (Hp:(rVlt p maxN))
  (Hp':(rVlt p' maxN))
  (Hq:(rVlt q maxN))
  (Hq':(rVlt q' maxN))
  (Hr:(rVlt r maxN))
  (Hr':(rVlt r' maxN))
  (Hs:(rVlt s maxN))
  (Hs':(rVlt s' maxN))
  p'=(evalMZ maxN M p Hp) ->
  q'=(evalMZ maxN M q Hq) ->
  r'=(evalMZ maxN M r Hr) ->
  s'=(evalMZ maxN M s Hs) ->
  (doTripletP (memF maxN M) t (addEq (p, q) (addEq (r, s) (memF maxN M)))) ->
  (ex
     State
     [S:State]
     (doTripletP (memF maxN M) t S) /\
     (eqState S (memFx (add2EqMem M p' q' r' s' Hp' Hq' Hr' Hs')))).
Intros M t p p' q q' r r' s s' Hp Hp' Hq Hq' Hr Hr' Hs Hs' H' H'0 H'1 H'2 H'3;
 Exists (addEq (p, q) (addEq (r, s) (memF maxN M))); Split; Auto.
Apply eqStateTrans
     with S2 :=
          (addEq
             (p, q)
             (addEq ((evalMZ maxN M r Hr), (evalMZ maxN M s Hs)) (memF maxN M)));
 Auto.
Apply eqStateAddEq; Auto.
Apply addEqComp; Auto; Apply evalEq; Auto.
Rewrite <- H'1; Rewrite <- H'2.
Apply eqStateTrans
     with S2 :=
          (addEq
             ((evalMZ maxN M p Hp), (evalMZ maxN M q Hq))
             (addEq (r', s') (memF maxN M))); Auto.
Apply addEqComp; Apply eqStateaddEq2; Apply evalEq; Auto.
Rewrite <- H'; Rewrite <- H'0.
Apply eqStateTrans
     with S2 := (addEq (p', q') (memFx (addEqMem maxN M r' s' Hr' Hs'))); Auto.
Apply eqStateAddEq; Auto.
Apply addCorrectExEq; Auto.
Unfold add2EqMem.
Case (addEqMem maxN M r' s' Hr' Hs'); Auto.
Intros p0; Case p0; Auto.
Intros m b; Case b; Simpl; Auto.
Intros H'0; Red; Split; Auto.
Intros d.
Generalize (addCorrectExEq m p' q' Hp' Hq'); Auto.
Simpl; Auto.
Case (addEqMem maxN m p' q' Hp' Hq').
Intros p1; Case p1; Auto.
Qed.
Hints Resolve doTripletFAux2.

Definition doTripletF:
  (M:(Mem maxN)) (t:triplet) (In t LL) ->{R:(Option ?) | Cases R of
                                                             (Some R') =>
                                                               (Ex [S:State]
                                                               (doTripletP
                                                                  (memF maxN M)
                                                                  t S) /\
                                                               (eqState
                                                                  S (memFx R')))
                                                               /\
                                                               ((eqExceptOn
                                                                   maxN M
                                                                   (getMem
                                                                      maxN R')
                                                                   (getList
                                                                      maxN R'))
                                                                /\
                                                                ((nEq
                                                                    maxN M
                                                                    (getMem
                                                                       maxN R')
                                                                    (getList
                                                                       maxN R'))
                                                                 /\
                                                                 (inclState
                                                                    (memF
                                                                       maxN M)
                                                                    (memF
                                                                       maxN
                                                                       (getMem
                                                                          maxN
                                                                          R')))))
                                                            | None =>
                                                                (S:State)
                                                                ~ (doTripletP
                                                                     (memF
                                                                        maxN M)
                                                                     t S)
                                                         end}.
Intros M t.
Case t; Intros b p q r Int.
Cut (rVlt p maxN);
 [Intros Hp |
   Apply inTripletmaxN with 1 := HmaxN;
    Apply inTripletsIn with t := (Triplet b p q r)]; Simpl; Auto.
Cut (rVlt q maxN);
 [Intros Hq |
   Apply inTripletmaxN with 1 := HmaxN;
    Apply inTripletsIn with t := (Triplet b p q r)]; Simpl; Auto.
Cut (rVlt r maxN);
 [Intros Hr |
   Apply inTripletmaxN with 1 := HmaxN;
    Apply inTripletsIn with t := (Triplet b p q r)]; Simpl; Auto.
Apply LetP with h := (evalMZ maxN M p Hp); Intros p' Eqp'.
Apply LetP with h := (evalMZ maxN M q Hq); Intros q' Eqq'.
Apply LetP with h := (evalMZ maxN M r Hr); Intros r' Eqr'.
Cut (rVlt p' maxN); [Intros Hp' | Rewrite Eqp'; Apply evalMZlt]; Auto.
Cut (rVlt q' maxN); [Intros Hq' | Rewrite Eqq'; Apply evalMZlt]; Auto.
Cut (rVlt r' maxN); [Intros Hr' | Rewrite Eqr'; Apply evalMZlt]; Auto.
Case b.
Case (rZDec p' (rZComp q')); Rewrite Eqp'; Rewrite Eqq'; Intros Eq1.
Exists (Some
          ?
          (add2EqMem
             M q' rZTrue r' rZFalse Hq' (MaxTrue LL maxN HmaxN) Hr'
             (MaxFalse LL maxN HmaxN))).
Split; Auto.
Apply doTripletFAux2
     with p := q q := rZTrue r := r s := rZFalse Hp := Hq
          Hq := (MaxTrue LL maxN HmaxN) Hr := Hr Hs := (MaxFalse LL maxN HmaxN);
 Auto.
Apply sym_equal; Apply evalMZTrue; Auto.
Apply sym_equal; Apply evalMZFalse; Auto.
Apply doTripletAndPpmq; Auto.
Apply evalInvComp with Maxa := Hp Maxb := Hq; Auto.
Case (rZDec p' (rZComp r')); Rewrite Eqp'; Rewrite Eqr'; Intros Eq2;
 Repeat Split; Auto.
Exists (Some
          ?
          (add2EqMem
             M q' rZFalse r' rZTrue Hq' (MaxFalse LL maxN HmaxN) Hr'
             (MaxTrue LL maxN HmaxN))).
Split; Auto; Auto.
Apply doTripletFAux2
     with p := q q := rZFalse r := r s := rZTrue Hp := Hq
          Hq := (MaxFalse LL maxN HmaxN) Hr := Hr Hs := (MaxTrue LL maxN HmaxN);
 Auto.
Apply sym_equal; Apply evalMZFalse; Auto.
Apply sym_equal; Apply evalMZTrue; Auto.
Apply doTripletAndPpmr; Auto.
Apply evalInvComp with Maxa := Hp Maxb := Hr; Auto.
Case (rZDec q' r'); Rewrite Eqq'; Rewrite Eqr'; Intros Eq3.
Exists (Some ? (addEqMem maxN M p' r' Hp' Hr')).
Split; Auto.
Apply doTripletFAux1 with p := p q := r Hp := Hp Hq := Hr; Auto.
Apply doTripletAndPqr; Auto.
Apply evalInv with Maxa := Hq Maxb := Hr; Auto.
Case (rZDec q' (rZComp r')); Rewrite Eqq'; Rewrite Eqr'; Intros Eq4.
Exists (Some ? (addEqMem maxN M p' rZFalse Hp' (MaxFalse LL maxN HmaxN))).
Split; Auto.
Apply doTripletFAux1
     with p := p q := rZFalse Hp := Hp Hq := (MaxFalse LL maxN HmaxN); Auto.
Apply sym_equal; Apply evalMZFalse; Auto.
Apply doTripletAndPqmr; Auto.
Apply evalInvComp with Maxa := Hq Maxb := Hr; Auto.
Case (rZDec p' rZTrue); Rewrite Eqp'; Rewrite <- (evalMZTrue LL maxN HmaxN M);
 Intros Eq5.
Exists (Some
          ?
          (add2EqMem
             M q' rZTrue r' rZTrue Hq' (MaxTrue LL maxN HmaxN) Hr'
             (MaxTrue LL maxN HmaxN))).
Split; Auto.
Apply doTripletFAux2
     with p := q q := rZTrue r := r s := rZTrue Hp := Hq
          Hq := (MaxTrue LL maxN HmaxN) Hr := Hr Hs := (MaxTrue LL maxN HmaxN);
 Auto.
Apply sym_equal; Apply evalMZTrue; Auto.
Apply sym_equal; Apply evalMZTrue; Auto.
Apply doTripletAndPpT; Auto.
Apply evalInv with Maxa := Hp Maxb := (MaxTrue LL maxN HmaxN); Auto.
Case (rZDec q' rZTrue); Rewrite Eqq'; Rewrite <- (evalMZTrue LL maxN HmaxN M);
 Intros Eq6.
Exists (Some ? (addEqMem maxN M p' r' Hp' Hr')).
Split; Auto.
Apply doTripletFAux1 with p := p q := r Hp := Hp Hq := Hr; Auto.
Apply doTripletAndPqT; Auto.
Apply evalInv with Maxa := Hq Maxb := (MaxTrue LL maxN HmaxN); Auto.
Case (rZDec q' rZFalse); Rewrite Eqq'; Rewrite <- (evalMZFalse LL maxN HmaxN M);
 Intros Eq7.
Exists (Some ? (addEqMem maxN M p' rZFalse Hp' (MaxFalse LL maxN HmaxN))).
Split; Auto.
Apply doTripletFAux1
     with p := p q := rZFalse Hp := Hp Hq := (MaxFalse LL maxN HmaxN); Auto.
Apply sym_equal; Apply evalMZFalse; Auto.
Apply doTripletAndPqF; Auto.
Apply evalInv with Maxa := Hq Maxb := (MaxFalse LL maxN HmaxN); Auto.
Case (rZDec r' rZTrue); Rewrite Eqr'; Rewrite <- (evalMZTrue LL maxN HmaxN M);
 Intros Eq8.
Exists (Some ? (addEqMem maxN M p' q' Hp' Hq')).
Split; Auto.
Apply doTripletFAux1 with p := p q := q Hp := Hp Hq := Hq; Auto.
Apply doTripletAndPrT; Auto.
Apply evalInv with Maxa := Hr Maxb := (MaxTrue LL maxN HmaxN); Auto.
Case (rZDec r' rZFalse); Rewrite Eqr'; Rewrite <- (evalMZFalse LL maxN HmaxN M);
 Intros Eq9.
Exists (Some ? (addEqMem maxN M p' rZFalse Hp' (MaxFalse LL maxN HmaxN))).
Split; Auto.
Apply doTripletFAux1
     with p := p q := rZFalse Hp := Hp Hq := (MaxFalse LL maxN HmaxN); Auto.
Apply sym_equal; Apply evalMZFalse; Auto.
Apply doTripletAndPrF; Auto.
Apply evalInv with Maxa := Hr Maxb := (MaxFalse LL maxN HmaxN); Auto.
Exists (None ((Mem maxN) * bool) * (dList maxN)).
Generalize evalCorrectComp; Generalize evalCorrect; Intros eV1 eV2; Intros S;
 Red; Intros H'; Inversion H'; Auto.
Case (rZDec p' q'); Rewrite Eqp'; Rewrite Eqq'; Intros Eq1.
Exists (Some ? (addEqMem maxN M r' rZTrue Hr' (MaxTrue LL maxN HmaxN))).
Split; Auto.
Apply doTripletFAux1
     with p := r q := rZTrue Hp := Hr Hq := (MaxTrue LL maxN HmaxN); Auto.
Apply sym_equal; Apply evalMZTrue; Auto.
Apply doTripletEqPpq; Auto.
Apply evalInv with 1 := Eq1; Auto.
Case (rZDec p' (rZComp q')); Rewrite Eqp'; Rewrite Eqq'; Intros Eq2.
Exists (Some ? (addEqMem maxN M r' rZFalse Hr' (MaxFalse LL maxN HmaxN))).
Split; Auto.
Apply doTripletFAux1
     with p := r q := rZFalse Hp := Hr Hq := (MaxFalse LL maxN HmaxN); Auto.
Apply sym_equal; Apply evalMZFalse; Auto.
Apply doTripletEqPpmq; Auto.
Apply evalInvComp with 1 := Eq2; Auto.
Case (rZDec p' r'); Rewrite Eqp'; Rewrite Eqr'; Intros Eq3.
Exists (Some ? (addEqMem maxN M q' rZTrue Hq' (MaxTrue LL maxN HmaxN))).
Split; Auto.
Apply doTripletFAux1
     with p := q q := rZTrue Hp := Hq Hq := (MaxTrue LL maxN HmaxN); Auto.
Apply sym_equal; Apply evalMZTrue; Auto.
Apply doTripletEqPpr; Auto.
Apply evalInv with 1 := Eq3; Auto.
Case (rZDec p' (rZComp r')); Rewrite Eqp'; Rewrite Eqr'; Intros Eq4.
Exists (Some ? (addEqMem maxN M q' rZFalse Hq' (MaxFalse LL maxN HmaxN))).
Split; Auto.
Apply doTripletFAux1
     with p := q q := rZFalse Hp := Hq Hq := (MaxFalse LL maxN HmaxN); Auto.
Apply sym_equal; Apply evalMZFalse; Auto.
Apply doTripletEqPpmr; Auto.
Apply evalInvComp with 1 := Eq4; Auto.
Case (rZDec q' r'); Rewrite Eqq'; Rewrite Eqr'; Intros Eq5.
Exists (Some ? (addEqMem maxN M p' rZTrue Hp' (MaxTrue LL maxN HmaxN))).
Split; Auto.
Apply doTripletFAux1
     with p := p q := rZTrue Hp := Hp Hq := (MaxTrue LL maxN HmaxN); Auto.
Apply sym_equal; Apply evalMZTrue; Auto.
Apply doTripletEqPqr; Auto.
Apply evalInv with 1 := Eq5; Auto.
Case (rZDec q' (rZComp r')); Rewrite Eqq'; Rewrite Eqr'; Intros Eq6.
Exists (Some ? (addEqMem maxN M p' rZFalse Hp' (MaxFalse LL maxN HmaxN))).
Split; Auto.
Apply doTripletFAux1
     with p := p q := rZFalse Hp := Hp Hq := (MaxFalse LL maxN HmaxN); Auto.
Apply sym_equal; Apply evalMZFalse; Auto.
Apply doTripletEqPqmr; Auto.
Apply evalInvComp with 1 := Eq6; Auto.
Case (rZDec p' rZTrue); Rewrite Eqp'; Rewrite <- (evalMZTrue LL maxN HmaxN M);
 Intros Eq7.
Exists (Some ? (addEqMem maxN M q' r' Hq' Hr')).
Split; Auto.
Apply doTripletFAux1 with p := q q := r Hp := Hq Hq := Hr; Auto.
Apply doTripletEqPpT; Auto.
Apply evalInv with 1 := Eq7; Auto.
Case (rZDec p' rZFalse); Rewrite Eqp'; Rewrite <- (evalMZFalse LL maxN HmaxN M);
 Intros Eq8.
Exists (Some ? (addEqMem maxN M q' (rZComp r') Hq' (rVltrZComp ? ? Hr'))).
Split; Auto.
Apply doTripletFAux1
     with p := q q := (rZComp r) Hp := Hq Hq := (rVltrZComp ? ? Hr); Auto.
Rewrite (evalMZComp maxN M r Hr); Auto.
Rewrite Eqr'; Auto.
Apply doTripletEqPpF; Auto.
Apply evalInv with 1 := Eq8; Auto.
Case (rZDec q' rZTrue); Rewrite Eqq'; Rewrite <- (evalMZTrue LL maxN HmaxN M);
 Intros Eq9.
Exists (Some ? (addEqMem maxN M p' r' Hp' Hr')).
Split; Auto.
Apply doTripletFAux1 with p := p q := r Hp := Hp Hq := Hr; Auto.
Apply doTripletEqPqT; Auto.
Apply evalInv with 1 := Eq9; Auto.
Case (rZDec q' rZFalse); Rewrite Eqq'; Rewrite <- (evalMZFalse LL maxN HmaxN M);
 Intros Eq10.
Exists (Some ? (addEqMem maxN M p' (rZComp r') Hp' (rVltrZComp ? ? Hr'))).
Split; Auto.
Apply doTripletFAux1
     with p := p q := (rZComp r) Hp := Hp Hq := (rVltrZComp ? ? Hr); Auto.
Rewrite (evalMZComp maxN M r Hr); Auto.
Rewrite Eqr'; Auto.
Apply doTripletEqPqF; Auto.
Apply evalInv with 1 := Eq10; Auto.
Case (rZDec r' rZTrue); Rewrite Eqr'; Rewrite <- (evalMZTrue LL maxN HmaxN M);
 Intros Eq11.
Exists (Some ? (addEqMem maxN M p' q' Hp' Hq')).
Split; Auto.
Apply doTripletFAux1 with p := p q := q Hp := Hp Hq := Hq; Auto.
Apply doTripletEqPrT; Auto.
Apply evalInv with 1 := Eq11; Auto.
Case (rZDec r' rZFalse); Rewrite Eqr'; Rewrite <- (evalMZFalse LL maxN HmaxN M);
 Intros Eq12.
Exists (Some ? (addEqMem maxN M p' (rZComp q') Hp' (rVltrZComp ? ? Hq'))).
Split; Auto.
Apply doTripletFAux1
     with p := p q := (rZComp q) Hp := Hp Hq := (rVltrZComp ? ? Hq); Auto.
Rewrite (evalMZComp maxN M q Hq); Auto.
Rewrite Eqq'; Auto.
Apply doTripletEqPrF; Auto.
Apply evalInv with 1 := Eq12; Auto.
Exists (None ((Mem maxN) * bool) * (dList maxN)).
Generalize evalCorrectComp; Generalize evalCorrect; Intros eV1 eV2; Intros S;
 Red; Intros H'; Inversion H'; Auto.
Defined.
End algo.
