Require Import PolyListAux.
Require Import makeTriplet.
Require Export stalmarck.

Definition TotalOn :=
   [S:State] [L:(list triplet)] (a:rZ) (In a (varTriplets L)) ->
   (eqStateRz S a rZTrue) \/ (eqStateRz S a rZFalse).

Lemma TotalProp1:
  (S:State) (L:(list triplet)) (TotalOn S L) ->
  (a, b:rZ) (In a (varTriplets L)) -> (In b (varTriplets L)) ->
  (eqStateRz S a b) \/ (eqStateRz S a (rZComp b)).
Unfold TotalOn; Intros S L H a b Ha Hb.
Elim (H a Ha); Elim (H b Hb); Intros H1 H2.
Left; Apply eqStateRzTrans with b := rZTrue; Auto.
Right; Apply eqStateRzTrans with b := rZTrue; Auto.
Cut rZTrue=(rZComp rZFalse); [Intros T; Rewrite T | Simpl]; Auto.
Right; Apply eqStateRzTrans with b := rZFalse; Auto.
Cut rZFalse=(rZComp rZTrue); [Intros T; Rewrite T | Simpl]; Auto.
Left; Apply eqStateRzTrans with b := rZFalse; Auto.
Qed.

Lemma TotalProp2:
  (S:State) (L:(list triplet)) (TotalOn S L) ->
  (a, b:rZ) (inTriplets a L) -> (inTriplets b L) ->
  (eqStateRz S a b) \/ (eqStateRz S a (rZComp b)).
Intros S L H a b Ha Hb.
Case (inTripletsVarTriplet L a Ha); Case (inTripletsVarTriplet L b Hb);
 Intros H1 H2.
Apply (TotalProp1 S L); Auto.
LApply (TotalProp1 S L H a (rZComp b) H2); Auto.
Rewrite (rZCompInv b); Auto; Intros Z; Elim Z; Simpl; Auto.
LApply (TotalProp1 S L H (rZComp a) b H2); Auto.
Intros Z; Elim Z; Intros Z1; Auto; Right; Rewrite <- (rZCompInv a); Auto.
Rewrite <- (rZCompInv a); Rewrite <- (rZCompInv b).
LApply (TotalProp1 S L H (rZComp a) (rZComp b) H2); Auto.
Intros Z; Elim Z; Simpl; Auto.
Qed.

Lemma TotalProp3:
  (S:State)
  (L:(list triplet)) (f:rNat ->bool) (TotalOn S L) -> (realizeState f S) ->
  (a, b:rZ) (inTriplets a L) -> (inTriplets b L) -> (rZEval f a)=(rZEval f b) ->
  (eqStateRz S a b).
Intros S L f T Hf a b Ha Hb; Elim (TotalProp2 S L T a b Ha Hb); Auto.
Intros H; Rewrite (realizeStateInv f S Hf a (rZComp b) H).
Rewrite (rZEvalCompInv b f).
Case (rZEval f b); Simpl; Intros H; Absurd true=false; Auto with bool.
Qed.

Lemma TotalProp4:
  (S:State)
  (L:(list triplet))
  (f, f':rNat ->bool)
  (TotalOn S L) ->
  (realizeState f S) ->
  (realizeState f' S) -> (f zero)=true -> (f' zero)=true ->
  (a:rZ) (inTriplets a L) ->(rZEval f a)=(rZEval f' a).
Intros S L f f' T R R' Z Z' a Ha; Elim (TotalProp2 S L T a rZTrue); Auto.
Intros H; Rewrite (realizeStateInv f' S R' a rZTrue); Auto;
 Rewrite (realizeStateInv f S R a rZTrue); Simpl; Auto; Rewrite Z; Auto.
Intros H; Rewrite (realizeStateInv f' S R' a (rZComp rZTrue)); Auto;
 Rewrite (realizeStateInv f S R a (rZComp rZTrue)); Simpl; Auto; Rewrite Z;
 Rewrite Z'; Auto.
Apply inTripletsTrue; Auto.
Qed.

Lemma TotalProp5:
  (S:State)
  (L:(list triplet))
  (f, f':rNat ->bool)
  (TotalOn S L) ->
  (realizeState f S) ->
  (realizeState f' S) ->
  (f zero)=true -> (f' zero)=true -> (realizeTriplets f L) ->
  (realizeTriplets f' L).
Unfold realizeTriplets; Unfold tEval; Intros S L f f' T R R' Z Z' H t; Case t;
 Intros n p q r Ht.
Rewrite <- (TotalProp4 S L f f' T R R' Z Z' p).
Rewrite <- (TotalProp4 S L f f' T R R' Z Z' q).
Rewrite <- (TotalProp4 S L f f' T R R' Z Z' r).
Apply (H (Triplet n p q r)); Auto.
Apply inTripletsIn with t := (Triplet n p q r); Auto; Unfold inTriplet; Auto.
Apply inTripletsIn with t := (Triplet n p q r); Auto; Unfold inTriplet; Auto.
Apply inTripletsIn with t := (Triplet n p q r); Auto; Unfold inTriplet; Auto.
Qed.

Theorem TotalDoTriplet:
  (S:State) (L:(list triplet)) (TotalOn S L) ->
  (t:triplet) (In t L) ->(ex
                            ? [u:rZ * rZ]Cases u of
                                             (pair a b) =>
                                               (inTriplets a L) /\
                                               (inTriplets b L)
                                         end /\ (doTripletP S t (addEq u S))).
Unfold TotalOn; Intros S L H t; Case t; Intros b p q r Ht.
Cut (inTriplets p L); [Intros Hp | Apply inTripletsIn with 2 := Ht]; Simpl; Auto.
Cut (inTriplets r L); [Intros Hr | Apply inTripletsIn with 2 := Ht]; Simpl; Auto.
Cut (inTriplets (rZComp r) L); [Intros Hr2 | Apply inTripletsIn with 2 := Ht];
 Simpl; Auto.
Cut (inTriplets rZFalse L); [Intros Hz | Apply inTripletsFalse]; Auto.
Elim (H q);
 [Intros Hq; Case b | Intros Hq; Case b | Apply varTripletTriplet2 with 1 := Ht];
 Auto.
Exists (p, r); Simpl; Split; [Split | Apply doTripletAndPqT]; Auto.
Exists (p, r); Simpl; Split; [Split | Apply doTripletEqPqT]; Auto.
Exists (p, rZFalse); Simpl; Split; [Split | Apply doTripletAndPqF]; Auto.
Exists (p, (rZComp r)); Simpl; Split; [Split | Apply doTripletEqPqF]; Auto.
Qed.

Lemma DoTripletAux:
  (S:State)
  (f:rNat ->bool)
  (t:triplet)
  (a, b:rZ)
  (doTripletP S t (addEq (a, b) S)) ->
  (realizeState f S) -> (f zero)=true -> (tEval f t)=false ->
  (rZEval f a)=(rZEval f (rZComp b)).
Intros S f t a b H1 H2 H3 H4; Rewrite (rZEvalCompInv b f).
Cut ~ (rZEval f a)=(rZEval f b); [Idtac | Red; Intros H].
Case (rZEval f a); Case (rZEval f b); Simpl; Auto with bool; Intros H; Elim H;
 Auto with bool.
Rewrite (evalRealizeState f S (addEq (a, b) S) t H1) in H4; Auto with bool.
Qed.

Definition RealizeTripletsDec:
  (f:rNat ->bool) (L:(list triplet))
  {(realizeTriplets f L)}+{~ (realizeTriplets f L)}.
Induction L.
Left; Auto.
Intros t L' HR.
Elim HR; Intros H.
Cut (tEval f t)=(tEval f t); Auto; Pattern 2 (tEval f t); Case (tEval f t).
Intros Ht; Left; Auto.
Intros Ht; Right; Unfold realizeTriplets; Red; Intros Abs.
Absurd (tEval f t)=true; Auto.
Rewrite Ht; Simpl; Auto with bool.
Apply Abs; Simpl; Auto.
Right; Red; Intros Abs.
Absurd (realizeTriplets f L'); Auto;
 Apply realizeTripletIncl with L1 := (cons t L'); Auto with datatypes.
Qed.

Lemma negRealizeTriplets:
  (f:rNat ->bool) (L:(list triplet)) ~ (realizeTriplets f L) ->
  (ex triplet [t:triplet](In t L) /\ (tEval f t)=false).
Induction L.
Intros Abs; Absurd (realizeTriplets f (nil triplet)); Auto.
Intros t L' HR Abs.
Cut (tEval f t)=(tEval f t); Auto; Pattern 2 (tEval f t); Case (tEval f t).
Intros Ht.
Case (RealizeTripletsDec f L'); Intros HL'.
Absurd (realizeTriplets f (cons t L')); Auto.
Elim (HR HL'); Intros t0 H; Elim H; Exists t0; Simpl; Auto.
Intros H; Exists t; Auto with datatypes.
Qed.

Definition eqStateRzContrDec2:
  (S:State){(contradictory S)}+{~ (contradictory S)}.
Intros S; Case (eqStateRzContrDec S); Intros A; Unfold contradictory;
 [Left | Right; Red; Intros B].
Exists (rZPlus zero); Auto.
Elim B; Intros a Ha; Auto.
Absurd (a, b:rZ)(eqStateRz S a b); Auto.
Intros c b; Apply (eqStateRzContr a); Auto.
Qed.

Definition DeductibleEq :=
   [L:(list triplet)]
   [S:State]
   [a, b:rZ]
   (f:rNat ->bool)
   (f zero)=true -> (realizeState f S) -> (realizeTriplets f L) ->
   (rZEval f a)=(rZEval f b).

Lemma DeductibleProp1:
  (L:(list triplet))
  (S, S':State) (a, b:rZ) (inclState S S') -> (DeductibleEq L S a b) ->
  (DeductibleEq L S' a b).
Intros L S S' a b H; Unfold DeductibleEq; Intros H' f F1 F2 F3; Apply H'; Auto.
Apply realizeStateIncl with S1 := S'; Auto.
Qed.

Lemma DeductibleProp2:
  (L:(list triplet))
  (S, S':State) (a, b:rZ) (dilemmaP S L S') -> (DeductibleEq L S' a b) ->
  (DeductibleEq L S a b).
Intros L S S' a b H; Unfold DeductibleEq; Intros H' f F1 F2 F3; Apply H'; Auto.
EApply dilemmaRealizeStateEval; EAuto.
Qed.

Lemma DeductibleProp3:
  (L:(list triplet)) (S:State) (a, b:rZ) (eqStateRz S a b) ->
  (DeductibleEq L S a b).
Intros L S a b H; Unfold DeductibleEq; Intros f F1 F2 F3; Auto.
EApply realizeStateInv; EAuto.
Qed.

Lemma DeductibleProp4:
  (L:(list triplet))
  (S:State) (a, b:rZ) (validEquation L a b) -> (eqStateRz S a (rZComp b)) ->
  (c, d:rZ)(DeductibleEq L S c d).
Unfold validEquation; Intros L S a b H H' c d; Unfold DeductibleEq;
 Intros f F1 F2 F3.
Absurd (rZEval f a)=(rZEval f (rZComp b));
 [Rewrite (H f F3 F1) | EApply realizeStateInv; EAuto].
Rewrite (rZEvalCompInv b f); Case (rZEval f b); Simpl; Auto with bool.
Qed.

Lemma DeductibleProp5:
  (L:(list triplet)) (S:State) (TotalOn S L) ->
  (f:rNat ->bool)
  (f zero)=true -> (realizeState f S) -> (realizeTriplets f L) ->
  (a, b:rZ) (DeductibleEq L S a b) ->(eqStateRz S a b).
Intros L S T f Hf1 Hf2 Hf3 a b H.
Apply realizeEqStateRz; Intros f' Hf'1 Hf'2.
Apply (H f'); Auto.
Apply (TotalProp5 S L f f'); Auto.
Qed.

Theorem DotripletsTotalDeductible:
  (S:State) (L:(list triplet)) (TotalOn S L) ->
  (ex
     State
     [S':State]
     (doTripletsP S L S') /\
     ((a, b:rZ) (DeductibleEq L S a b) ->(eqStateRz S' a b))).
Intros S L T; Unfold DeductibleEq.
Case (eqStateRzContrDec2 S); Intros HS.
Exists S; Split; Auto; Intros a b Hab; Auto.
Unfold contradictory in HS; Elim HS; Intros c Hc; EApply eqStateRzContr; EAuto.
Elim (Realizable2 S HS); Intros f Hf; Elim Hf; Intros Hf1 Hf2.
Case (RealizeTripletsDec f L); Intros HL.
Exists S; Split; Auto; Intros a b Hab; Auto.
EApply DeductibleProp5; EAuto.
Elim (negRealizeTriplets f L); [Intros t Ht; Elim Ht; Intros Ht1 Ht2 | Auto].
Elim (TotalDoTriplet S L T t Ht1); Intros p; Case p; Intros c d U; Elim U;
 Intros U1 U2; Elim U1; Intros Hc Hd.
Exists (addEq (c, d) S); Split; Auto.
Apply doTripletsTrans with t := t S2 := (addEq (c, d) S); Auto.
Intros a b Hab; Apply eqStateRzContr with a := d.
Apply eqStateRzTrans with b := c; Auto.
Cut (eqStateRz S c (rZComp d)); Auto.
Apply (TotalProp3 S L f T Hf1 c (rZComp d)); Auto.
Apply inTripletsComp; Auto.
Apply (DoTripletAux S f t); Auto.
Qed.
Require Import Arith.

Fixpoint nthTail[n:nat]: (list rZ) ->(list rZ) :=
   [l:(list rZ)]
      Cases n l of
          O l => l
         | (S n') nil => (nil ?)
         | (S n') (cons t l') => (nthTail n' l')
      end.

Fixpoint myNth[n:nat; l:(list rZ)]: rZ ->rZ :=
   [default:?]
      Cases n l of
          O (cons x l') => x
         | (S m) (cons x t) => (myNth m t default)
         | _ _ => default
      end.

Lemma nthTailProp1:
  (n:nat) (l:(list rZ)) (lt n (length l)) ->
  (nthTail n l)=(cons (myNth n l rZTrue) (nthTail (S n) l)).
Induction n.
Intros l; Case l; Simpl; Intros; Auto; Absurd (lt O O); Auto with arith.
Intros n' HR l; Case l; Simpl;
 [Intros H; Absurd (lt (S n') O) | Intros t l' H; Rewrite (HR l')];
 Auto with arith.
Qed.

Lemma nthTailProp2:
  (n:nat) (l:(list rZ)) (le (length l) n) ->(nthTail n l)=(nil ?).
Induction n.
Intros l; Case l; Simpl; Auto; Intros t l' H; Absurd (le (S (length l')) O);
 Auto with arith.
Intros n' HR l; Case l; Simpl; Auto.
Intros t l' H; Rewrite (HR l'); Auto with arith.
Qed.

Lemma nthTailProp3:
  (n:nat) (l:(list rZ)) (le (length l) n) ->(nthTail n l)=(nthTail (S n) l).
Intros n l H; Rewrite (nthTailProp2 n l H); Apply sym_eq; Apply nthTailProp2;
 Auto with arith.
Qed.

Lemma DilemmaTotalDeductibleAux:
  (n:nat)
  (S:State)
  (L:(list triplet))
  ((a:rZ) (In a (nthTail n (varTriplets L))) ->
   (eqStateRz S a rZTrue) \/ (eqStateRz S a rZFalse)) ->
  (ex
     State
     [S':State]
     (dilemmaP S L S') /\
     ((a, b:rZ) (DeductibleEq L S a b) ->(eqStateRz S' a b))).
Induction n.
Simpl; Intros S L H; Elim (DotripletsTotalDeductible S L H); Intros S' HS'.
Elim HS'; Intros HS'1 HS'2; Exists S'; Auto.
Intros n' HR S L H.
Elim (le_or_lt (length (varTriplets L)) n'); Intros Hn'.
Apply (HR S L); Auto.
Rewrite (nthTailProp3 n' (varTriplets L)); Auto.
Elim (HR (addEq ((myNth n' (varTriplets L) rZTrue), rZTrue) S) L); Auto.
Intros S1 HS1; Elim HS1; Intros HS11 HS12.
Elim (HR (addEq ((myNth n' (varTriplets L) rZTrue), rZFalse) S) L); Auto.
Intros S2 HS2; Elim HS2; Intros HS21 HS22.
Exists (interState S1 S2); Split; Auto.
Apply dilemmaPSplit
     with a := (myNth n' (varTriplets L) rZTrue) b := rZTrue S2 := S1 S3 := S2;
 Auto.
Intros a b Hab; Apply interMemEqStateRz.
Apply HS12; Apply (DeductibleProp1 L S); Auto.
Apply HS22; Apply (DeductibleProp1 L S); Auto.
Rewrite (nthTailProp1 n' (varTriplets L)); Auto; Simpl; Intros c Hc; Elim Hc;
 Intros Hc'.
Rewrite Hc'; Right; Auto.
Elim (H c); [Left | Right | Idtac]; Auto.
Rewrite (nthTailProp1 n' (varTriplets L)); Auto; Simpl; Intros c Hc; Elim Hc;
 Intros Hc'.
Rewrite Hc'; Left; Auto.
Elim (H c); [Left | Right | Idtac]; Auto.
Qed.

Theorem DilemmaTotalDeductible:
  (S:State) (L:(list triplet))
  (ex
     State
     [S':State]
     (dilemmaP S L S') /\
     ((a, b:rZ) (DeductibleEq L S a b) ->(eqStateRz S' a b))).
Intros S L; Apply (DilemmaTotalDeductibleAux (length (varTriplets L)) S L); Auto.
Rewrite (nthTailProp2 (length (varTriplets L)) (varTriplets L)); Auto.
Intros a H; Inversion H.
Qed.

Theorem DilemmaConfluenceAux:
  (L:(list triplet)) (S1:State)
  (ex State [S2:State] (S3:State) (dilemmaP S1 L S3) ->(dilemmaP S3 L S2)).
Intros L S1; Elim (DilemmaTotalDeductible S1 L); Intros S2 HS2; Exists S2.
Intros S3 HS3; Elim (DilemmaTotalDeductible S3 L); Intros S4 HS4.
Elim HS2; Intros HS2_1 HS2_2; Elim HS4; Intros HS4_1 HS4_2.
Apply dilemmaComp with S1 := S3 S2 := S4; Auto.
Unfold eqState; Split; Unfold inclState; Intros a b Hab.
Apply HS4_2; Apply (DeductibleProp1 L S1); Auto.
EApply dilemmaIncl; EAuto.
Apply (DeductibleProp2 L S1 S2); Auto; Apply DeductibleProp3; Auto.
Apply HS2_2; Apply (DeductibleProp2 L S1 S3); Auto.
Apply (DeductibleProp2 L S3 S4); Auto; Apply DeductibleProp3; Auto.
Qed.

Theorem DilemmaConfluence:
  (L:(list triplet)) (S, S1, S2:State) (dilemmaP S L S1) -> (dilemmaP S L S2) ->
  (ex State [S3:State](dilemmaP S1 L S3) /\ (dilemmaP S2 L S3)).
Intros L S S1 S2 HS1 HS2; Elim (DilemmaConfluenceAux L S); Intros S0 HS0.
Exists S0; Split; Auto.
Qed.

Theorem StalmarckConfluenceAux:
  (L:(list triplet)) (S, S1:State) (stalmarckP S L S1) ->
  (S2:State) (dilemmaP S L S2) ->
  (ex State [S3:State](dilemmaP S1 L S3) /\ (stalmarckP S2 L S3)).
Intros L S S1 H'; Elim H'; Auto.
Intros S2 S3 L0 H'0 S4 H'1; Exists S4; Split; Auto.
Apply dilemmaComp with S1 := S2 S2 := S4; Auto.
Intros S2 S3 S4 L0 H'0 H'1 H'2 S5 H'3.
Elim (DilemmaConfluence L0 S2 S5 S3);
 [Intros S6 E; Elim E; Intros H'10 H'11; Clear E | Idtac | Idtac]; Auto.
Elim (H'2 S6); [Intros S7 E; Elim E; Intros H'6 H'7; Clear E | Idtac]; Auto.
Exists S7; Split; Auto.
Apply stalmarckTrans with S2 := S6; Auto.
Qed.

Theorem StalmarckConfluence:
  (L:(list triplet))
  (S, S1, S2:State) (stalmarckP S L S1) -> (stalmarckP S L S2) ->
  (ex State [S3:State](stalmarckP S1 L S3) /\ (stalmarckP S2 L S3)).
Intros L S S1 S2 H'; Generalize S2; Elim H'; Auto.
Intros S3 S4 L0 H'0 S5 H'1; Exists S5; Split; Auto.
Apply stalmarckComp with S1 := S3 S2 := S5; Auto.
Intros S3 S4 S5 L0 H'0 H'1 H'2 S6 H'3.
LApply (StalmarckConfluenceAux L0 S3 S6);
 [Intro H'7; Elim (H'7 S4);
   [Intros S7 E; Elim E; Intros H'10 H'11; Clear E | Idtac] | Idtac]; Auto.
Elim (H'2 S7); [Intros S9 E; Elim E; Intros H'6 H'8; Clear E | Idtac]; Auto.
Exists S9; Split; Auto.
Apply stalmarckTrans with S2 := S7; Auto.
Qed.

Theorem dilemmaComplete:
  (e:Expr) (Tautology e) ->Cases (makeTriplets (norm e)) of
                               (tRC l s n) =>
                                 (ex
                                    State
                                    [S:State]
                                    (dilemmaP (addEq (s, rZFalse) (nil ?)) l S)
                                    /\ (contradictory S))
                           end.
Intros e; Generalize (refl_equal ? (makeTriplets (norm e)));
 Pattern -2 (makeTriplets (norm e)); Case (makeTriplets (norm e)).
Intros l r r0 H H1.
Elim (DilemmaTotalDeductible (addEq (r, rZFalse) (nil ?)) l).
Intros S HS; Exists S; Auto.
Elim HS; Intros HS1 HS2; Split; Auto.
Elim (TautoRTauto e); Intros Thm1 Thm2.
Elim (rTautotTauto (norm e)); Intros Thm3 Thm4.
Unfold tTautology in Thm3; Rewrite <- H in Thm3; Simpl; Auto.
Unfold contradictory.
Exists rZTrue; Apply HS2;
 Apply (DeductibleProp4
          l (addEq <rZ, rZ> (r, rZFalse) (nil rZ * rZ)) r (rZPlus zero)); Auto.
Qed.

Theorem stalmarckComplete:
  (e:Expr) (Tautology e) ->Cases (makeTriplets (norm e)) of
                               (tRC l s n) =>
                                 (ex
                                    State
                                    [S:State]
                                    (stalmarckP
                                       (addEq (s, rZFalse) (nil ?)) l S) /\
                                    (contradictory S))
                           end.
Intros e H'; Generalize (dilemmaComplete e H'); Case (makeTriplets (norm e)).
Intros l r H'0 H'1; Elim H'1; Intros S E; Elim E; Intros H'2 H'3; Clear E H'1.
Exists S; Split; Auto.
Apply stalmarckTrans with S2 := S; Auto.
Qed.

