
(* Definition of the dilemma rule *)
Require Export doTriplets.
Require Export interState.

Mutual Inductive dilemmaP: State -> (list triplet) -> State ->Prop :=
     dilemmaPRef:
       (S1, S2:State) (L:(list triplet)) (doTripletsP S1 L S2) ->
       (dilemmaP S1 L S2)
    | dilemmaPSplit:
        (a, b:rZ)
        (S1, S2, S3, S4:State)
        (L:(list triplet))
        (dilemmaP (addEq (a, b) S1) L S2) ->
        (dilemmaP (addEq (a, (rZComp b)) S1) L S3) ->
        (eqState (interState S2 S3) S4) ->(dilemmaP S1 L S4) .
Hints Resolve dilemmaPRef.

Definition boolDec: (a, b:bool){a=b}+{~ a=b}.
Intros a b; Case a; Case b; Auto; Right; Red; Intros; Discriminate.
Defined.

Theorem dilemmaComp:
  (S1, S2, S3, S4:State)
  (L:(list triplet)) (dilemmaP S1 L S2) -> (eqState S3 S1) -> (eqState S4 S2) ->
  (dilemmaP S3 L S4).
Intros S1 S2 S3 S4 L H'; Generalize S3 S4; Elim H'; Clear H' L S1 S2 S3 S4.
Intros S1 S2 L H' S3 S4 H'0 H'1; Apply dilemmaPRef; Auto.
Apply doTripletsComp with S1 := S1 S2 := S2; Auto.
Intros a b S1 S2 S3 S4 L H' H'0 H'1 H'2 H'3 S0 S5 H'4 H'5.
Apply dilemmaPSplit with a := a b := b S2 := S2 S3 := S3; Auto.
Apply eqStateTrans with S2 := S4; Auto.
Qed.

Theorem dilemmaUnionEx:
  (S1, S2:State) (L:(list triplet)) (dilemmaP S1 L S2) ->
  (Ex [S3:State] (eqState S2 (unionState S3 S1))).
Intros S1 S2 L H'; Elim H'; Clear H' L S1 S2; Auto.
Exact doTripletsUnionEx.
Intros a b S1 S2 S3 S4 L H' H'0 H'1 H'2 H'3.
Exists S4; Auto.
Red; Split; Auto.
Apply unionStateMin; Auto.
Apply inclStateEqStateComp with S1 := S1 S3 := (interState S2 S3); Auto.
Apply interMemMin; Auto.
Elim H'0; Intros S5 E; Clear H'0.
Apply inclStateEqStateComp
     with S1 := S1 S3 := (unionState (unionState S5 (addEq (a, b) (nil ?))) S1);
 Auto.
Apply eqStateTrans
     with S2 := (unionState S5 (unionState (addEq (a, b) (nil rZ * rZ)) S1));
 Auto.
Apply eqStateSym; Auto.
Apply unionStateAssoc; Auto.
Elim H'2; Intros S5 E; Clear H'0.
Apply inclStateEqStateComp
     with S1 := S1
          S3 := (unionState (unionState S5 (addEq (a, (rZComp b)) (nil ?))) S1);
 Auto.
Apply eqStateTrans
     with S2 :=
          (unionState S5 (unionState (addEq (a, (rZComp b)) (nil rZ * rZ)) S1));
 Auto.
Apply eqStateSym; Auto.
Apply unionStateAssoc; Auto.
Qed.

Theorem dilemmaIncl:
  (S1, S2:State) (L:(list triplet)) (dilemmaP S1 L S2) ->(inclState S1 S2).
Intros S1 S2 L H'.
Elim (dilemmaUnionEx S1 S2 L); [Intros S3 E | Idtac]; Auto.
Apply inclStateEqStateComp with S1 := S1 S3 := (unionState S3 S1); Auto.
Qed.

Theorem dilemmaMonotoneEx:
  (S1, S2, S3:State)
  (L:(list triplet)) (dilemmaP S1 L S3) -> (inclState S1 S2) ->
  (Ex [S4:State] (dilemmaP S2 L S4) /\ (inclState S3 S4)).
Intros S1 S2 S3 L H'; Generalize S2; Elim H'; Auto; Clear H' L S1 S2 S3.
Intros S1 S3 L H' S2 H'0.
Elim (doTripletsMonotoneEx S1 S2 S3 L);
 [Intros S4 E; Elim E; Intros H'7 H'8; Clear E | Idtac | Idtac]; Auto.
Exists S4; Split; Auto.
Intros a b S1 S2 S3 S4 L H' H'0 H'1 H'2 H'3 S0 H'4.
Elim (H'0 (addEq (a, b) S0));
 [Intros S5 E; Elim E; Intros H'7 H'8; Clear E | Idtac]; Auto.
Elim (H'2 (addEq (a, (rZComp b)) S0));
 [Intros S6 E; Elim E; Intros H'9 H'10; Clear E | Idtac]; Auto.
Exists (interState S5 S6); Split; Auto.
Apply dilemmaPSplit with 1 := H'7 2 := H'9; Auto.
Apply inclStateEqStateComp
     with S1 := (interState S2 S3) S3 := (interState S5 S6); Auto.
Apply interMemMin; Auto.
Apply inclStateTrans with S2 := S2; Auto.
Apply inclStateTrans with S2 := S3; Auto.
Qed.
(* We don't loose realizability of states  if the triplets are realized *)

Theorem dilemmaRealizeStateEval:
  (f:rNat ->bool)
  (S1, S2:State)
  (L:(list triplet))
  (realizeState f S1) ->
  (dilemmaP S1 L S2) -> (realizeTriplets f L) -> (f zero)=true ->
  (realizeState f S2).
Intros f S1 S2 L H' H'0; Generalize H'; Elim H'0; Auto.
Intros S3 S4 L0 H'1 H'2 H'3 H'4.
Apply doTripletsRealizeStateEval with S1 := S3 L := L0; Auto.
Intros a b S3 S4 S5 S6 L0 H'1 H'2 H'3 H'4 H'5 H'6 H'7 H'8.
Case (boolDec (rZEval f a) (rZEval f b)); Intros Eqf.
Apply realizeStateIncl with S1 := S4; Auto.
Inversion H'5; Auto.
Apply inclStateTrans with S2 := (interState S4 S5); Auto.
Apply realizeStateIncl with S1 := S5; Auto.
Apply H'4; Auto.
Apply realizeStateAddEq; Auto.
Rewrite rZEvalCompInv.
Generalize Eqf; Case (rZEval f a); Case (rZEval f b); Simpl; Auto; Intros H'10;
 Case H'10; Auto.
Apply inclStateEqStateComp with S1 := (interState S4 S5) S3 := S5; Auto.
Qed.
