(* Definition of equality on booleans upto a certain number *)

Require Import Relation_Definitions.
Require Export triplet.

Definition equalBefore :=
   [n:rNat] [f, g:rNat ->bool] (m:rNat) (rlt m n) ->(f m)=(g m).

Lemma equalBeforeElim:
  (n, m:rNat) (f, g:rNat ->bool) (equalBefore n f g) -> (rlt m n) ->(f m)=(g m).
Intros n m f g H' H'1; Red in H'; Auto.
Qed.

Lemma equalBeforeTrans: (n:rNat)(transitive rNat ->bool (equalBefore n)).
Intros n; Red; Auto.
Intros x y z H' H'0; Red; Auto.
Intros m H'1; Apply trans_eq with y := (y m); Auto.
Qed.

Lemma equalBeforeLt:
  (n, m:rNat) (f, g:rNat ->bool) (rlt n (rnext m)) -> (equalBefore m f g) ->
  (equalBefore n f g).
Intros n m f g H' H'0; Red; Red in H'0; Auto.
Intros m0 H'1; Apply H'0; Auto.
Apply rltTransRnext2 with m := n; Auto.
Qed.

Lemma equalBeforeSym: (n:rNat)(symmetric rNat ->bool (equalBefore n)).
Intros n; Red.
Intros x y H'; Red in H'; Red; Auto.
Intros m H'0; Apply sym_eq; Auto.
Qed.
Hints Resolve equalBeforeSym.

Definition extendFun :=
   [n:rNat] [g:rNat ->bool] [s:bool] [m:rNat]
      Cases (rltDec m n) of
          (inleft (left _)) => (g m)
         | _ => s
      end.

Lemma equalBeforeExtend:
  (g:rNat ->bool) (n, m:rNat) (s:bool) (rlt n (rnext m)) ->
  (equalBefore n g (extendFun m g s)).
Intros g n m s H; Simpl; Auto.
Red; Auto.
Intros p H'; Unfold extendFun; Case (rltDec p m); Auto.
Intros s0; Case s0; Auto.
Intros H'0; Absurd (rlt (rnext m) n); Auto.
Apply rltAntiSym; Auto.
Apply rltTransRnext1 with m := p; Auto.
Intros H'0; Rewrite <- H'0 in H.
Case (rNextInv n p); Auto.
Intros H'1; Absurd p=n; Auto.
Apply rltDef2; Auto.
Intros H'1; Absurd (rlt n p); Auto.
Apply rltAntiSym; Auto.
Qed.
Hints Resolve equalBeforeExtend.

Lemma extendFunrZEvalExact:
  (g:rNat ->bool) (n:rNat) (p:rZ) (s:bool) (valRz p)=n ->
  (rZEval (extendFun n g s) p)=(Cases p of
                                    (rZPlus p) => s
                                   | (rZMinus _) => (negb s)
                                end).
Intros g n p; Case p; Unfold extendFun; Simpl; Auto; Intros r s H;
 Case (rltDec r n); Auto; Intros tmp; Case tmp; Rewrite H; Intros;
 Absurd (rlt n n); Auto.
Qed.

Lemma extendFunrZEval:
  (g:rNat ->bool) (n:rNat) (p:rZ) (s:bool) (rVlt p n) ->
  (rZEval (extendFun n g s) p)=(rZEval g p).
Intros g n p; Case p; Unfold rVlt; Simpl; Auto.
Intros r s H'.
Unfold extendFun.
Case (rltDec r n); Auto.
Intros s0; Case s0; Auto.
Intros H'0; Absurd (rlt n r); Auto.
Apply rltAntiSym; Auto.
Intros H'0; Absurd r=n; Auto.
Apply rltDef2; Auto.
Intros r s H'.
Unfold extendFun.
Case (rltDec r n); Auto.
Intros s0; Case s0; Auto.
Intros H'0; Absurd (rlt n r); Auto.
Apply rltAntiSym; Auto.
Intros H'0; Absurd r=n; Auto.
Apply rltDef2; Auto.
Qed.

Lemma equalBeforerZEval:
  (f, g:rNat ->bool) (n:rNat) (p:rZ) (rVlt p n) -> (equalBefore n f g) ->
  (rZEval f p)=(rZEval g p).
Intros f g n p; Case p; Unfold rVlt; Simpl; Auto.
Intros r H' H'0; Red in H'0; Auto.
Rewrite (H'0 r); Auto.
Qed.

Theorem equalBeforeNext:
  (f, g:rNat ->bool) (n:rNat) (equalBefore n f g) -> (f n)=(g n) ->
  (equalBefore (rnext n) f g).
Intros f g n H' H'0; Red in H'; Red; Auto.
Intros m H'1; Case (rNextInv m n); Auto.
Intros H'2; Rewrite H'2; Auto.
Qed.

Lemma equalBeforeTEval:
  (f, g:rNat ->bool) (t:triplet) (equalBefore (rnext (maxVarTriplet t)) f g) ->
  (tEval f t)=(tEval g t).
Intros f g t; Case t; Simpl; Auto.
Intros r r0 r1 r2 H'; Red in H'.
Cut (rZEval f r0)=(rZEval g r0); [Intros Eq1; Rewrite Eq1 | Idtac]; Auto.
Cut (rZEval f r1)=(rZEval g r1); [Intros Eq2; Rewrite Eq2 | Idtac]; Auto.
Cut (rZEval f r2)=(rZEval g r2); [Intros Eq3; Rewrite Eq3 | Idtac]; Auto.
Generalize H'; Case r2; Simpl; Auto; Intros r3 H'0; Rewrite H'0; Simpl; Auto.
Generalize H'; Case r1; Simpl; Auto; Intros r3 H'0; Rewrite H'0; Simpl; Auto.
Apply rltTransRnext1 with m := (rmax (valRz r0) r3); Auto.
Apply rltTransRnext1 with m := (rmax (valRz r0) r3); Auto.
Generalize H'; Case r0; Simpl; Auto; Intros r3 H'0; Rewrite H'0; Simpl; Auto.
Apply rltTransRnext1 with m := (rmax r3 (valRz r1)); Auto.
Apply rltTransRnext1 with m := (rmax r3 (valRz r1)); Auto.
Qed.
Hints Resolve equalBeforeTEval.

Lemma supportTriplets:
  (f, g:rNat ->bool)
  (l:(list triplet))
  (realizeTriplets f l) -> (equalBefore (rnext (maxVarTriplets l)) f g) ->
  (realizeTriplets g l).
Intros f g l; Elim l; Simpl; Auto.
Intros a l0 H' H'0 H'1.
Red; Simpl; Auto.
Intros t H'2; Elim H'2;
 [Intros H'3; Rewrite <- H'3; Clear H'2 | Intros H'3; Clear H'2]; Auto.
Apply trans_eq with y := (tEval f a); Auto.
Apply sym_eq.
Apply equalBeforeTEval; Auto.
Apply equalBeforeLt with 2 := H'1; Auto.
Red in H'0; Auto with datatypes.
LApply H';
 [Intro H'2; LApply H'2; [Intro H'4; Clear H'2 H' | Clear H'2 H'] | Clear H'];
 Auto.
Apply equalBeforeLt with 2 := H'1; Auto.
Red; Red in H'0; Auto with datatypes.
Qed.
Hints Resolve supportTriplets.

Theorem inLt: (e:rExpr) (n:rNat) (inRExpr n e) ->(rlt n (rnext (maxVar e))).
Induction e; Intros.
Inversion H.
Inversion H.
Inversion H; Simpl; Auto.
Inversion H0; Simpl; Auto.
Inversion H1; Simpl; Auto.
Apply rltTransRnext1 with m := (maxVar r0); Auto.
Apply rltTransRnext1 with m := (maxVar r1); Auto.
Qed.

Theorem equalBeforeREval:
  (f, g:rNat ->bool) (e:rExpr) (equalBefore (rnext (maxVar e)) f g) ->
  (rEval f e)=(rEval g e).
Intros f g e H'.
Apply support; Auto.
Intros n H'0.
Apply equalBeforeElim with n := (rnext (maxVar e)); Auto.
Apply inLt; Auto.
Qed.

