(* Definition of the transformation from rExpr to list of triplets *)

Require Export equalBefore.

Mutual Inductive tripletResult: Set :=
     tRC: (list triplet) -> rZ -> rNat ->tripletResult .

Fixpoint makeTripletsFun[l:(list triplet); n:rNat; e:rExpr]: tripletResult :=
   Cases e of
       (rV i) => (tRC l (rZPlus i) n)
      | (rN p) => Cases (makeTripletsFun l n p) of
                      (tRC l1 s1 n1) => (tRC l1 (rZComp s1) n1)
                  end
      | (rNode node p q) => Cases (makeTripletsFun l n p) of
                                (tRC l1 s1 n1) =>
                                  Cases (makeTripletsFun l1 n1 q) of
                                      (tRC l2 s2 n2) =>
                                        (tRC
                                           (cons
                                              (Triplet node (rZPlus n2) s1 s2)
                                              l2) (rZPlus n2) (rnext n2))
                                  end
                            end
   end.

Definition makeTriplets: rExpr ->tripletResult :=
   [e:rExpr](makeTripletsFun (nil ?) (rnext (maxVar e)) e).

Lemma makeTripletsFunMax:
  (e:rExpr)
  (l, l1:(list triplet))
  (n, n1:rNat)
  (s1:rZ) (rlt (maxVar e) n) -> (makeTripletsFun l n e)=(tRC l1 s1 n1) ->
  (rVlt s1 n1).
Intros e; Elim e; Simpl; Auto.
Intros r H' l1 n n1 s1 H'0 H'1; Inversion H'1; Simpl; Auto.
Rewrite <- H2; Auto.
Intros r H' l l1 n n1 s1 H'0.
Generalize (refl_equal ? (makeTripletsFun l n r));
 Pattern 2 3 (makeTripletsFun l n r); Case (makeTripletsFun l n r); Auto.
Intros l0 r0 r1 H'1 H'2.
LApply (H' l l0 n r1 r0);
 [Intros H'8; LApply H'8; [Intros H'9; Clear H'8 | Clear H'8] | Idtac]; Auto.
Generalize H'9; Inversion H'2; Simpl; Auto.
Intros r r0 H' r1 H'0 l l1 n n1 s1 H'1.
Case (makeTripletsFun l n r0); Auto.
Intros l0 r2 r3.
Case (makeTripletsFun l0 r3 r1); Auto.
Intros l2 r4 r5 H'2; Inversion H'2; Simpl; Auto.
Red; Auto.
Qed.

Lemma makeTripletsFunIncr:
  (e:rExpr)
  (l, l':(list triplet))
  (n, n':rNat) (s':rZ) (makeTripletsFun l n e)=(tRC l' s' n') ->
  (rlt n (rnext n')).
Intros e; Elim e; Simpl; Auto.
Intros r H' l' n n' s' H'0; Inversion H'0; Auto.
Intros r H' l l' n n' s'.
Generalize (refl_equal ? (makeTripletsFun l n r));
 Pattern 2 3 (makeTripletsFun l n r); Case (makeTripletsFun l n r); Auto.
Intros l0 r0 r1 H'0 H'1; LApply (H' l l0 n r1 r0); Auto; Inversion H'1; Auto.
Intros r r0 H' r1 H'0 l l' n n' s'.
Generalize (refl_equal ? (makeTripletsFun l n r0));
 Pattern 2 3 (makeTripletsFun l n r0); Case (makeTripletsFun l n r0); Auto.
Intros l0 r2 r3 H'1.
Generalize (refl_equal ? (makeTripletsFun l0 r3 r1));
 Pattern 2 3 (makeTripletsFun l0 r3 r1); Case (makeTripletsFun l0 r3 r1); Auto.
Intros l1 r4 r5 H'2 H'3.
LApply (H' l l0 n r3 r2); [Intros H'9 | Idtac]; Auto.
LApply (H'0 l0 l1 r3 r5 r4); [Intros H'10 | Idtac]; Auto.
Inversion H'3; Auto.
Apply rltTrans with y := (rnext r3); Auto.
Qed.
Hints Resolve makeTripletsFunIncr.

Lemma makeTripletsFunIncl:
  (e:rExpr)
  (l, l1:(list triplet))
  (n, n1:rNat) (s1:rZ) (makeTripletsFun l n e)=(tRC l1 s1 n1) ->(incl l l1).
Intros e; Elim e; Simpl.
Intros r l l1 n n1 s1 H'; Inversion H'; Auto with datatypes.
Intros r H' l l1 n n1 s1.
Generalize (refl_equal ? (makeTripletsFun l n r));
 Pattern 2 3 (makeTripletsFun l n r); Case (makeTripletsFun l n r).
Intros l0 r0 r1 H'0 H'1.
LApply (H' l l0 n r1 r0); [Intros H'7 | Idtac]; Auto.
Inversion H'1; Auto.
Rewrite <- H0; Auto.
Intros r r0 H' r1 H'0 l l1 n n1 s1.
Generalize (refl_equal ? (makeTripletsFun l n r0));
 Pattern 2 3 (makeTripletsFun l n r0); Case (makeTripletsFun l n r0).
Intros l0 r2 r3 H'1.
Generalize (refl_equal ? (makeTripletsFun l0 r3 r1));
 Pattern 2 3 (makeTripletsFun l0 r3 r1); Case (makeTripletsFun l0 r3 r1).
Intros l2 r4 r5 H'2 H'3.
LApply (H' l l0 n r3 r2); [Intros H'9 | Idtac]; Auto.
LApply (H'0 l0 l2 r3 r5 r4); [Intros H'10 | Idtac]; Auto.
Apply incl_tran with l0; Auto.
Apply incl_tran with l2; Auto.
Inversion H'3; Auto with datatypes.
Qed.
Hints Resolve makeTripletsFunIncr.

Lemma maxVarTripletsRlt:
  (e:rExpr)
  (l, l':(list triplet))
  (n, n':rNat)
  (s':rZ)
  (rlt (maxVar e) n) ->
  (rlt (maxVarTriplets l) n) -> (makeTripletsFun l n e)=(tRC l' s' n') ->
  (rlt (maxVarTriplets l') n').
Intros e; Elim e; Simpl.
Intros r l l' n n' s' H' H'0 H'1; Generalize H'0; Inversion H'1; Auto.
Intros r H' l l' n n' s' H'0 H'1.
Generalize (refl_equal ? (makeTripletsFun l n r));
 Pattern 2 3 (makeTripletsFun l n r); Case (makeTripletsFun l n r).
Intros l0 r0 r1 H'2 H'3.
LApply (H' l l0 n r1 r0);
 [Intros H'9; LApply H'9;
   [Intros H'10; LApply H'10; [Clear H'10 H'9 | Clear H'10 H'9] | Clear H'9] |
   Idtac]; Auto.
Inversion H'3; Auto.
Intros r r0 H' r1 H'0 l l' n n' s' H'1 H'2.
Generalize (refl_equal ? (makeTripletsFun l n r0));
 Pattern 2 3 (makeTripletsFun l n r0); Case (makeTripletsFun l n r0).
Intros l0 r2 r3 H'3.
Generalize (refl_equal ? (makeTripletsFun l0 r3 r1));
 Pattern 2 3 (makeTripletsFun l0 r3 r1); Case (makeTripletsFun l0 r3 r1).
Intros l1 r4 r5 H'4 H'5.
Inversion H'5; Simpl; Auto.
Apply rmaxRlt; Auto.
Apply rmaxRlt; Auto.
Apply rmaxRlt; Auto.
Apply rltTrans with y := r3; Auto.
Change (rVlt r2 r3).
Apply makeTripletsFunMax with e := r0 l := l l1 := l0 n := n; Auto.
Apply rmaxRltLeft with p := (maxVar r1); Auto.
Apply makeTripletsFunIncr with e := r1 l := l0 l' := l1 s' := r4; Auto.
Apply rltTrans with y := r5; Auto.
Change (rVlt r4 r5).
Apply makeTripletsFunMax with e := r1 l := l0 l1 := l1 n := r3; Auto.
Apply rltTransRnext2 with m := n; Auto.
Apply rmaxRltRight with m := (maxVar r0); Auto.
Apply makeTripletsFunIncr with e := r0 l := l l' := l0 s' := r2; Auto.
Apply rltTrans with y := r5; Auto.
Apply H'0 with l := l0 n := r3 s' := r4; Auto.
Apply rltTransRnext2 with m := n; Auto.
Apply rmaxRltRight with m := (maxVar r0); Auto.
Apply makeTripletsFunIncr with e := r0 l := l l' := l0 s' := r2; Auto.
Apply H' with l := l n := n s' := r2; Auto.
Apply rmaxRltLeft with p := (maxVar r1); Auto.
Qed.

Theorem extendEvalMakeTripletsFun:
  (e:rExpr)
  (f:rNat ->bool)
  (l, l':(list triplet))
  (n, n':rNat)
  (s':rZ)
  (rlt (maxVar e) n) ->
  (rlt (maxVarTriplets l) n) ->
  (realizeTriplets f l) -> (makeTripletsFun l n e)=(tRC l' s' n') ->
  (Ex [g:rNat ->bool]
  (equalBefore n f g) /\ ((realizeTriplets g l') /\ (rEval f e)=(rZEval g s'))).
Intros e; Elim e; Simpl.
Intros r f l l' n n' s' H' H'0 H'1 H'2; Inversion H'2; Simpl.
Exists f; Split; [Idtac | Split]; Auto; Try Rewrite <- H0; Auto.
Red; Auto.
Intros r H' f l l' n n' s' H'0 H'1 H'2.
Generalize (refl_equal ? (makeTripletsFun l n r));
 Pattern 2 3 (makeTripletsFun l n r); Case (makeTripletsFun l n r).
Intros l0 r0 r1 H'3 H'4.
Elim (H' f l l0 n r1 r0);
 [Intros g E; Elim E; Intros H'15 H'16; Elim H'16; Intros H'17 H'18;
   Clear H'16 E | Idtac | Idtac | Idtac | Idtac]; Auto.
Inversion H'4; Rewrite <- H0; Rewrite H'18; Auto.
Exists g; Split; [Idtac | Split]; Simpl; Auto.
Case r0; Simpl; Auto.
Intros r2; Case (g r2); Auto.
Intros r r0 H' r1 H'0 f l l' n n' s' H'1 H'2 H'3.
Generalize (refl_equal ? (makeTripletsFun l n r0));
 Pattern 2 3 (makeTripletsFun l n r0); Case (makeTripletsFun l n r0).
Intros l0 r2 r3 H'4.
Generalize (refl_equal ? (makeTripletsFun l0 r3 r1));
 Pattern 2 3 (makeTripletsFun l0 r3 r1); Case (makeTripletsFun l0 r3 r1).
Intros l1 r4 r5 H'5 H'6.
Cut (rlt (maxVar r0) n);
 [Intros Rlt1 | Apply rmaxRltLeft with p := (maxVar r1)]; Auto.
Cut (rlt (maxVar r1) n);
 [Intros Rlt2 | Apply rmaxRltRight with m := (maxVar r0)]; Auto.
Cut (rlt n (rnext r3));
 [Intros Rlt3 | Apply makeTripletsFunIncr with e := r0 l := l l' := l0 s' := r2];
 Auto.
Cut (rlt (rmax (maxVar r0) (maxVar r1)) r3);
 [Intros Rllt0 | Apply rltTransRnext2 with m := n]; Auto.
Elim (H' f l l0 n r3 r2);
 [Intros g E; Elim E; Intros H'17 H'18; Elim H'18; Intros H'19 H'20;
   Clear H'18 E | Idtac | Idtac | Idtac | Idtac]; Auto.
Elim (H'0 g l0 l1 r3 r5 r4);
 [Intros g0 E; Elim E; Intros H'18 H'21; Elim H'21; Intros H'22 H'23;
   Clear H'21 E | Idtac | Idtac | Idtac | Idtac]; Auto.
2:Apply rltTransRnext2 with m := n; Auto.
2:Apply maxVarTripletsRlt with 3 := H'4; Auto.
Inversion H'6.
Cut (rlt r3 (rnext r5));
 [Intros Rlt4 | Apply makeTripletsFunIncr with 1 := H'5]; Auto.
Exists (extendFun r5 g0 (rBoolOpFun r (rEval f r0) (rEval g r1))); Split;
 [Idtac | Split]; Auto.
Apply (equalBeforeTrans n) with y := g; Auto.
Apply equalBeforeLt with m := r3; Auto.
Apply (equalBeforeTrans r3) with y := g0; Auto.
Red; Simpl; Auto.
Intros t H'7; Elim H'7;
 [Intros H'8; Rewrite <- H'8; Clear H'7 | Intros H'8; Clear H'7]; Auto.
Unfold tEval.
Rewrite (extendFunrZEval g0 r5 r2 (rBoolOpFun r (rEval f r0) (rEval g r1)));
 Auto.
Rewrite (extendFunrZEval g0 r5 r4 (rBoolOpFun r (rEval f r0) (rEval g r1)));
 Auto.
Rewrite extendFunrZEvalExact; Auto.
Rewrite H'20; Rewrite H'23; Auto; Auto.
Rewrite <- (equalBeforerZEval g g0 r3 r2); Auto.
Case (rBoolOpFun r (rZEval g r2) (rZEval g0 r4)); Auto.
Apply makeTripletsFunMax with e := r0 l := l l1 := l0 n := n; Auto.
Apply makeTripletsFunMax with e := r1 l := l0 l1 := l1 n := r3; Auto.
Apply rmaxRltRight with m := (maxVar r0); Auto.
Red.
Apply rltTransRnext2 with m := r3; Auto.
Change (rVlt r2 r3).
Apply makeTripletsFunMax with e := r0 l := l l1 := l0 n := n; Auto.
Cut (realizeTriplets
       (extendFun r5 g0 (rBoolOpFun r (rEval f r0) (rEval g r1))) l1); Auto.
Apply supportTriplets with f := g0; Auto.
Apply equalBeforeExtend; Auto.
Apply rnextMono; Auto.
Apply maxVarTripletsRlt with 3 := H'5; Auto.
Apply rmaxRltRight with m := (maxVar r0); Auto.
Apply maxVarTripletsRlt with 3 := H'4; Auto.
Rewrite extendFunrZEvalExact; Auto.
Rewrite (equalBeforeREval f g r1); Auto.
Apply equalBeforeLt with m := n; Auto.
Qed.

Theorem extendEvalMakeTriplets:
  (f:rNat ->bool)
  (e:rExpr) (l:(list triplet)) (n:rNat) (s:rZ) (makeTriplets e)=(tRC l s n) ->
  (Ex [g:rNat ->bool]
  (equalBefore (rnext (maxVar e)) f g) /\
  ((realizeTriplets g l) /\ (rEval f e)=(rZEval g s))).
Intros f e l n s H'.
Apply extendEvalMakeTripletsFun with l := (nil triplet) n' := n; Simpl; Auto.
Qed.

Theorem equalBeforeMakeTripletsFun:
  (e:rExpr)
  (f, g:rNat ->bool)
  (l, l':(list triplet))
  (n, n':rNat)
  (s':rZ)
  (rlt (maxVar e) n) ->
  (rlt (maxVarTriplets l) n) ->
  (equalBefore n f g) ->
  (realizeTriplets f l') ->
  (realizeTriplets g l') -> (makeTripletsFun l n e)=(tRC l' s' n') ->
  (equalBefore n' f g).
Intros e; Elim e; Simpl.
Intros r f g l l' n n' s' H' H'0 H'1 H'2 H'3 H'4; Inversion H'4; Rewrite <- H2;
 Auto.
Intros r H' f g l l' n n' s' H'0 H'1 H'2 H'3 H'4.
Generalize (refl_equal ? (makeTripletsFun l n r));
 Pattern 2 3 (makeTripletsFun l n r); Case (makeTripletsFun l n r); Auto.
Intros l0 r0 r1 H'5 H'6; Inversion H'6; Auto; Rewrite <- H2.
Apply H' with l := l l' := l0 n := n s' := r0; Auto; Rewrite H0; Auto.
Intros r r0 H' r1 H'0 f g l l' n n' s' H'1 H'2 H'3 H'4 H'5.
Generalize (refl_equal ? (makeTripletsFun l n r0));
 Pattern 2 3 (makeTripletsFun l n r0); Case (makeTripletsFun l n r0); Auto.
Intros l0 r2 r3 H'6.
Generalize (refl_equal ? (makeTripletsFun l0 r3 r1));
 Pattern 2 3 (makeTripletsFun l0 r3 r1); Case (makeTripletsFun l0 r3 r1); Auto.
Intros l1 r4 r5 H'7 H'8.
Inversion H'8.
Cut (rlt n (rnext r3)); [Intros Rle0 | Apply makeTripletsFunIncr with 1 := H'6];
 Auto.
Cut (equalBefore r5 f g); [Intros eB0 | Idtac]; Auto.
Apply equalBeforeNext; Auto.
Cut (tEval f (Triplet r (rZPlus r5) r2 r4))=true; [Simpl | Idtac].
Cut (tEval g (Triplet r (rZPlus r5) r2 r4))=true; [Simpl | Idtac].
Rewrite (equalBeforerZEval f g r5 r4); Auto.
Rewrite (equalBeforerZEval f g r5 r2); Auto.
Case (f r5); Auto; Case (g r5); Auto;
 Case (rBoolOpFun r (rZEval g r2) (rZEval g r4)); Auto.
Red.
Apply rltTransRnext2 with m := r3; Auto.
Change (rVlt r2 r3).
Apply makeTripletsFunMax with 2 := H'6; Auto.
Apply rmaxRltLeft with p := (maxVar r1); Auto.
Apply makeTripletsFunIncr with 1 := H'7; Auto.
Apply makeTripletsFunMax with 2 := H'7; Auto.
Apply rmaxRltRight with m := (maxVar r0); Auto.
Apply rltTransRnext2 with m := n; Auto.
Rewrite <- H0 in H'5; Red in H'5; Auto with datatypes.
Rewrite <- H0 in H'4; Red in H'4; Auto with datatypes.
Apply H'0 with 6 := H'7; Auto.
Apply rmaxRltRight with m := (maxVar r0); Auto.
Apply rltTransRnext2 with m := n; Auto.
Apply maxVarTripletsRlt with 3 := H'6; Auto.
Apply rmaxRltLeft with p := (maxVar r1); Auto.
Apply H' with 6 := H'6; Auto.
Apply rmaxRltLeft with p := (maxVar r1); Auto.
Apply realizeTripletIncl with L1 := l'; Auto.
Rewrite <- H0; Auto.
Apply incl_tran with m := l1; Auto with datatypes.
Apply makeTripletsFunIncl with 1 := H'7; Auto.
Apply realizeTripletIncl with L1 := l'; Auto.
Rewrite <- H0; Auto.
Apply incl_tran with m := l1; Auto with datatypes.
Apply makeTripletsFunIncl with 1 := H'7; Auto.
Apply realizeTripletIncl with L1 := l'; Auto.
Rewrite <- H0; Auto with datatypes.
Apply realizeTripletIncl with L1 := l'; Auto.
Rewrite <- H0; Auto with datatypes.
Qed.

Theorem equalBeforeMakeTriplets:
  (f, g:rNat ->bool)
  (e:rExpr)
  (l:(list triplet))
  (n:rNat)
  (s:rZ)
  (equalBefore (rnext (maxVar e)) f g) ->
  (realizeTriplets f l) ->
  (realizeTriplets g l) -> (makeTriplets e)=(tRC l s n) ->
  (rZEval f s)=(rZEval g s).
Intros f g e l n s H' H'0 H'1 H'2.
Apply equalBeforerZEval with n := n; Auto.
Apply makeTripletsFunMax
     with e := e l := (nil triplet) l1 := l n := (rnext (maxVar e)); Auto.
Apply equalBeforeMakeTripletsFun
     with e := e l := (nil triplet) l' := l n := (rnext (maxVar e)) s' := s;
 Auto.
Qed.

Theorem rZEvalREvalMakeTriplets:
  (f, g:rNat ->bool)
  (e:rExpr)
  (l:(list triplet))
  (n:rNat)
  (s:rZ)
  (equalBefore (rnext (maxVar e)) f g) ->
  (realizeTriplets g l) -> (makeTriplets e)=(tRC l s n) ->
  (rZEval g s)=(rEval f e).
Intros f g e l n s H' H'0 H'1.
Elim (extendEvalMakeTriplets f e l n s);
 [Intros g0 E; Elim E; Intros H'8 H'9; Elim H'9; Intros H'10 H'11; Rewrite H'11;
   Clear H'9 E | Idtac]; Auto.
Apply equalBeforeMakeTriplets with e := e l := l n := n; Auto.
Apply (equalBeforeTrans (rnext (maxVar e))) with y := f; Auto.
Apply (equalBeforeSym (rnext (maxVar e))); Auto.
Qed.

Theorem rZEvalEvalRZMakeTriplets:
  (f:rNat ->bool)
  (e:rExpr)
  (l:(list triplet))
  (n:rNat) (s:rZ) (realizeTriplets f l) -> (makeTriplets e)=(tRC l s n) ->
  (rEval f e)=(rZEval f s).
Intros f e l n s H' H'0.
Apply sym_equal; Auto.
Apply rZEvalREvalMakeTriplets with l := l n := n; Auto.
Red; Auto.
Qed.

Definition tTautology :=
   [e:rExpr]
      Cases (makeTriplets e) of
          (tRC l s n) => (validEquation l s (rZPlus zero))
      end.

Theorem rTautotTauto: (e:rExpr)(iff (rTautology e) (tTautology e)).
Intros e; Unfold tTautology; Unfold rTautology;
 Generalize (refl_equal ? (makeTriplets e)); Pattern 2 3 (makeTriplets e);
 Case (makeTriplets e).
Intros l r r0 H'; Split; Intros H'1.
Red; Simpl; Auto.
Intros f H'0 fZ0; Simpl.
Rewrite fZ0; Auto.
Rewrite <- (H'1 f); Auto.
Apply rZEvalREvalMakeTriplets with l := l n := r0; Auto.
Red; Auto.
Intros f fZ0.
Red in H'1.
Elim (extendEvalMakeTriplets f e l r0 r);
 [Intros g E; Elim E; Intros H'7 H'8; Elim H'8; Intros H'9 H'10; Clear H'8 E |
   Idtac]; Auto.
Rewrite H'10.
Rewrite (H'1 g); Simpl; Auto.
Rewrite <- equalBeforeElim with 1 := H'7; Auto.
Rewrite <- equalBeforeElim with 1 := H'7; Auto.
Qed.



Theorem makeTripletsIn: (e:rExpr)Cases (makeTriplets e) of
                                     (tRC l s n) =>
                                       ~ l=(nil ?) ->(inTriplets s l)
                                 end.
Unfold makeTriplets; Intros e; Elim e; Simpl; Auto.
Intros r H'; Case H'; Auto.
Intros r; Case (makeTripletsFun (nil triplet) (rnext (maxVar r)) r);
 Auto.
Intros l r0 H' H'0 H'1.
Apply inTripletsComp; Auto.
Intros r r0 H' r1 H'0.
Case (makeTripletsFun
        (nil triplet) (rnext (rmax (maxVar r0) (maxVar r1))) r0).
Intros l r2 r3.
Case (makeTripletsFun l r3 r1); Simpl; Auto.
Qed.

