
(* Definition of stalmarck as the transitive closure of the dilemma rule *)
Require Export dilemma.
Require Import makeTriplet.

Mutual Inductive stalmarckP: State -> (list triplet) -> State ->Prop :=
     stalmarckPref:
       (S1, S2:State) (L:(list triplet)) (eqState S1 S2) ->(stalmarckP S1 L S2)
    | stalmarckTrans:
        (S1, S2, S3:State)
        (L:(list triplet)) (dilemmaP S1 L S2) -> (stalmarckP S2 L S3) ->
        (stalmarckP S1 L S3) .
Hints Resolve stalmarckPref.

Theorem stalmarckrTrans:
  (S1, S2, S3:State)
  (L:(list triplet)) (stalmarckP S1 L S2) -> (stalmarckP S2 L S3) ->
  (stalmarckP S1 L S3).
Intros S1 S2 S3 L H'; Generalize S3; Elim H'; Clear S3 H'; Auto.
Intros S3 S4 L0 H' S5 H'0.
Apply stalmarckTrans with S2 := S4; Auto.
Intros S3 S4 S5 L0 H' H'0 H'1 S0 H'2.
Apply stalmarckTrans with S2 := S4; Auto.
Qed.

Theorem stalmarckComp:
  (S1, S2, S3, S4:State)
  (L:(list triplet))
  (stalmarckP S1 L S2) -> (eqState S3 S1) -> (eqState S4 S2) ->
  (stalmarckP S3 L S4).
Intros S1 S2 S3 S4 L H'; Generalize S3 S4; Elim H'; Clear H' L S1 S2 S3 S4.
Intros S1 S2 L H' S3 S4 H'0 H'1.
Apply stalmarckPref; Auto.
Apply eqStateTrans with S2 := S1; Auto.
Apply eqStateTrans with S2 := S2; Auto.
Intros S1 S2 S3 L H' H'0 H'1 S0 S4 H'2 H'3.
Apply stalmarckTrans with S2 := S2; Auto.
Apply dilemmaComp with S1 := S1 S2 := S2; Auto.
Qed.

Theorem stalmarckUnionEx:
  (S1, S2:State) (L:(list triplet)) (stalmarckP S1 L S2) ->
  (Ex [S3:State] (eqState S2 (unionState S3 S1))).
Intros S1 S2 L H'; Elim H'; Clear H' L S1 S2; Auto.
Intros S1 S2 H' H'0; Exists S2; Auto.
Apply eqStateTrans with S2 := (unionState S2 S2); Auto.
Red; Split; Auto.
Intros S1 S2 S3 L H' H'0 H'1; Elim H'1; Intros S0 E; Clear H'1.
Elim (dilemmaUnionEx S1 S2 L); [Intros S4 E0 | Idtac]; Auto.
Exists (unionState S0 S4); Auto.
Apply eqStateTrans with S2 := (unionState S0 S2); Auto.
Apply eqStateTrans with S2 := (unionState S0 (unionState S4 S1)); Auto.
Apply unionStateAssoc; Auto.
Qed.

Theorem stalmarckIncl:
  (S1, S2:State) (L:(list triplet)) (stalmarckP S1 L S2) ->(inclState S1 S2).
Intros S1 S2 L H'.
Elim (stalmarckUnionEx S1 S2 L); [Intros S3 E | Idtac]; Auto.
Apply inclStateEqStateComp with S1 := S1 S3 := (unionState S3 S1); Auto.
Qed.

Theorem stalmarckMonotoneEx:
  (S1, S2, S3:State)
  (L:(list triplet)) (stalmarckP S1 L S3) -> (inclState S1 S2) ->
  (Ex [S4:State] (stalmarckP S2 L S4) /\ (inclState S3 S4)).
Intros S1 S2 S3 L H'; Generalize S2; Elim H'; Auto; Clear H' L S1 S2 S3.
Intros S1 S3 L H' S2 H'0.
Exists S2; Split; Auto.
Apply inclStateEqStateComp with S1 := S1 S3 := S2; Auto.
Intros S1 S2 S3 L H' H'0 H'1 S0 H'2.
Elim (dilemmaMonotoneEx S1 S0 S2 L);
 [Intros S4 E; Elim E; Intros H'9 H'10; Clear E | Idtac | Idtac]; Auto.
Elim (H'1 S4); [Intros S5 E; Elim E; Intros H'5 H'6; Clear E | Idtac]; Auto.
Exists S5; Split; Auto.
Apply stalmarckTrans with S2 := S4; Auto.
Qed.
(* We don't loose realizability of states if the triplets are realized *)

Theorem stalmarckRealizeStateEval:
  (f:rNat ->bool)
  (S1, S2:State)
  (L:(list triplet))
  (realizeState f S1) ->
  (stalmarckP S1 L S2) -> (realizeTriplets f L) -> (f zero)=true ->
  (realizeState f S2).
Intros f S1 S2 L H' H'0; Generalize H'; Elim H'0; Auto.
Intros S3 S4 L0 H'1 H'2 H'3 H'4.
Apply realizeStateIncl with S1 := S3; Auto.
Inversion H'1; Auto.
Intros S3 S4 S5 L0 H'1 H'2 H'3 H'4 H'5 H'6.
Apply H'3; Auto.
Apply dilemmaRealizeStateEval with S1 := S3 L := L0; Auto.
Qed.

Theorem stalmarckGivesValidEquation:
  (L:(list triplet))
  (a, b:rZ)
  (S:State)
  (stalmarckP (addEq (a, (rZComp b)) (nil ?)) L S) -> (contradictory S) ->
  (validEquation L a b).
Intros L a b S H' H'0.
Red; Auto.
Intros f H'1 H'2; Case (boolDec (rZEval f a) (rZEval f b)); Auto.
Intros H'3; Absurd (realizeState f S); Auto.
Apply stalmarckRealizeStateEval
     with S1 := (addEq (a, (rZComp b)) (nil rZ * rZ)) L := L; Auto.
Apply realizeStateAddEq; Auto.
Rewrite rZEvalCompInv; Generalize H'3; Case (rZEval f a); Case (rZEval f b);
 Auto; Intros H'4; Case H'4; Auto.
Qed.

Theorem stalmarckCorrect: (e:Expr) (S:State)Cases (makeTriplets (norm e)) of
                                                (tRC l s n) =>
                                                  (stalmarckP
                                                     (addEq
                                                        (s, rZFalse) (nil ?)) l
                                                     S) -> (contradictory S) ->
                                                  (Tautology e)
                                            end.
Intros e S; Generalize (refl_equal ? (makeTriplets (norm e)));
 Pattern -2 (makeTriplets (norm e)); Case (makeTriplets (norm e)).
Intros l r r0 H' H'0 H'1.
Specialize 1 TautoRTauto with e := e; Intros H'2; Inversion H'2.
Apply H0; Auto.
Specialize 1 rTautotTauto with e := (norm e); Intros H'3; Inversion H'3.
Apply H2; Auto.
Red; Auto.
Rewrite <- H'; Auto.
Apply stalmarckGivesValidEquation with S := S; Auto.
Qed.
