
(* Definition of trace to represent stalmarck's computation *)
Require Export stalmarck.

Mutual Inductive Trace: Set :=
     emptyTrace: Trace
    | tripletTrace: triplet ->Trace
    | seqTrace: Trace -> Trace ->Trace
    | dilemmaTrace: rZ -> rZ -> Trace -> Trace ->Trace .

Mutual Inductive evalTrace: State -> Trace -> State ->Prop :=
     emptyTraceEval:
       (S1, S2:State) (eqState S1 S2) ->(evalTrace S1 emptyTrace S2)
    | tripletTraceEval:
        (t:triplet) (S1, S2:State) (doTripletP S1 t S2) ->
        (evalTrace S1 (tripletTrace t) S2)
    | seqTraceEval:
        (t1, t2:Trace)
        (S1, S2, S3:State) (evalTrace S1 t1 S2) -> (evalTrace S2 t2 S3) ->
        (evalTrace S1 (seqTrace t1 t2) S3)
    | dliemmaTraceEval:
        (t1, t2:Trace)
        (a, b:rZ)
        (S1, S2, S3, S4:State)
        (evalTrace (addEq (a, b) S1) t1 S2) ->
        (evalTrace (addEq (a, (rZComp b)) S1) t2 S3) ->
        (eqState (interState S2 S3) S4) ->
        (evalTrace S1 (dilemmaTrace a b t1 t2) S4) .
Hints Resolve emptyTraceEval tripletTraceEval.

Theorem evalTraceEq:
  (S1, S2:State) (t:Trace) (evalTrace S1 t S2) ->
  (S3, S4:State) (evalTrace S3 t S4) -> (eqState S1 S3) ->(eqState S2 S4).
Intros S1 S2 t H'; Elim H'; Clear H' t S2 S1; Auto.
Intros S1 S2 H' S3 S4 H'0; Inversion H'0; Auto.
Intros H'1.
Apply eqStateTrans with S2 := S1; Auto.
Apply eqStateTrans with S2 := S3; Auto.
Intros t S1 S2 H' S3 S4 H'0; Inversion H'0; Auto.
Intros H'1.
Apply doTripletEqComp with t := t S1 := S1 S2 := S3; Auto.
Intros t1 t2 S1 S2 S3 H' H'0 H'1 H'2 S0 S4 H'3; Inversion H'3; Auto.
Intros H'4.
Apply H'2 with S0 := S6; Auto.
Apply H'0 with S3 := S0; Auto.
Intros t1 t2 a b S1 S2 S3 S4 H' H'0 H'1 H'2 H'3 S0 S5 H'4; Inversion H'4; Auto.
Intros H'5.
Apply eqStateTrans with S2 := (interState S7 S8); Auto.
Apply eqStateTrans with S2 := (interState S2 S3); Auto.
Apply interStateEq; Auto.
Apply H'0 with S3 := (addEq (a, b) S0); Auto.
Apply H'2 with S0 := (addEq (a, (rZComp b)) S0); Auto.
Qed.

Fixpoint TraceInList[T:Trace]: (list triplet) ->Prop :=
   [L:(list triplet)]
      Cases T of
          emptyTrace => True
         | (tripletTrace t) => (In t L)
         | (seqTrace T1 T2) => (TraceInList T1 L) /\ (TraceInList T2 L)
         | (dilemmaTrace a b T1 T2) => (TraceInList T1 L) /\ (TraceInList T2 L)
      end.

Theorem doTripletsExTrace:
  (L:(list triplet)) (S1, S2:State) (doTripletsP S1 L S2) ->
  (Ex [t:Trace] (evalTrace S1 t S2) /\ (TraceInList t L)).
Intros L S1 S2 H'; Elim H'; Clear H' S1 S2 L.
Intros S1 S2 H' H'0; Exists emptyTrace; Simpl; Auto.
Intros S1 S2 S3 L t H' H'0 H'1 H'2; Elim H'2; Intros t0 E; Elim E;
 Intros H'3 H'4; Clear E H'2.
Exists (seqTrace (tripletTrace t) t0); Simpl; Repeat Split; Auto.
Apply seqTraceEval with S2 := S2; Auto.
Qed.

Theorem dilemmaExTrace:
  (L:(list triplet)) (S1, S2:State) (dilemmaP S1 L S2) ->
  (Ex [t:Trace] (evalTrace S1 t S2) /\ (TraceInList t L)).
Intros L S1 S2 H'; Elim H'; Clear H' S1 S2 L; Auto.
Intros S1 S2 L H'.
Apply doTripletsExTrace with L := L; Auto.
Intros a b S1 S2 S3 S4 L H' H'0 H'1 H'2 H'3.
Elim H'0; Intros t E; Elim E; Intros H'4 H'5; Clear E H'0.
Elim H'2; Intros t0 E; Elim E; Intros H'0 H'6; Clear E H'2.
Exists (dilemmaTrace a b t t0); Simpl; Repeat Split; Auto.
Apply dliemmaTraceEval with S2 := S2 S3 := S3; Auto.
Qed.

Theorem stalmarckExTrace:
  (L:(list triplet)) (S1, S2:State) (stalmarckP S1 L S2) ->
  (Ex [t:Trace] (evalTrace S1 t S2) /\ (TraceInList t L)).
Intros L S1 S2 H'; Elim H'; Clear H' S2 S1 L; Auto.
Intros S1 H'; Exists emptyTrace; Simpl; Auto.
Intros S1 S2 S3 L H' H'0 H'1; Elim H'1; Intros t E; Elim E; Intros H'2 H'3;
 Clear E H'1.
Elim (dilemmaExTrace L S1 S2);
 [Intros t0 E; Elim E; Intros H'7 H'8; Clear E | Idtac]; Auto.
Exists (seqTrace t0 t); Simpl; Repeat Split; Auto.
Apply seqTraceEval with S2 := S2; Auto.
Qed.
